////////////////////////////////////////////////////////////////////////////////////////
// File: GPUGeomDiff.cpp
//
// GPU implementation of the technique presented in the paper "Estimating Curvatures
// and their Derivatives on Meshes of Arbitrary Topology from Sampling Directions"
//
// Copyright (C) 2007 Harlen Costa Batagelo.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// Lesser General Public License for more details.
////////////////////////////////////////////////////////////////////////////////////////

#include "dxstdafx.h"
#include "resource.h"

#include <vector>
#include <algorithm>
using std::find;
using namespace std;

//#define DEBUG_VS   // Uncomment this line to debug vertex shaders 
//#define DEBUG_PS   // Uncomment this line to debug pixel shaders 

////////////////////////////////////////////////////////////////////////////////////////
// Visualization modes
////////////////////////////////////////////////////////////////////////////////////////
enum VISMODE
{
	CURVCOLOR,     // Minimum and maximum curvatures as colors
    GAUSSIANCURV,  // Gaussian curvature
    MEANCURV,      // Mean curvature
    PRINCIPALDIRS, // Principal directions
    MAGCURVDEV,    // Curvature derivative magnitude
	NONE
};

////////////////////////////////////////////////////////////////////////////////////////
// Mesh modes
////////////////////////////////////////////////////////////////////////////////////////
enum MESHES
{
	TEAPOT,
	KNOT,
	HEAD,
	BUNNY,
	HORSE,
	ARMADILLO
};

MESHES g_eCurMesh = TEAPOT; // Current mesh

const TCHAR g_szMesh[6][32] = 
{
	L"Teapot.x",
	L"Knot.x",
	L"Head.x",
	L"Bunny.x",
	L"Horse.x",
	L"Armadillo.x"
};

////////////////////////////////////////////////////////////////////////////////////////
// Vertex formats
////////////////////////////////////////////////////////////////////////////////////////
#define FVF_TLVERTEX ( D3DFVF_XYZRHW | D3DFVF_TEX1 | D3DFVF_TEXCOORDSIZE2(0) )
struct TLVERTEX { D3DXVECTOR4 p; D3DXVECTOR2 t; };
#define FVF_MESHVERTEX ( D3DFVF_XYZ | D3DFVF_NORMAL | D3DFVF_TEX1 | D3DFVF_TEXCOORDSIZE1(0) )
struct MESHVERTEX { D3DXVECTOR3 p; D3DXVECTOR3 n; FLOAT id; };

////////////////////////////////////////////////////////////////////////////////////////
// Global variables
////////////////////////////////////////////////////////////////////////////////////////
LPDIRECT3DDEVICE9          g_pd3dDevice     = NULL; // Pointer to the D3D device
LPDIRECT3DSURFACE9         g_pd3dBackBuffer = NULL; // Pointer to the back buffer surface
LPD3DXFONT                 g_pFont          = NULL; // Font for drawing text
LPD3DXSPRITE               g_pTextSprite    = NULL; // Sprite for batching draw text calls
LPD3DXEFFECT               g_pEffect        = NULL; // D3DX effect interface
LPD3DXMESH                 g_pMesh          = NULL; // D3DX original mesh
LPDIRECT3DVERTEXBUFFER9    g_pVB            = NULL; // Indexed VB of the mesh to render
LPDIRECT3DINDEXBUFFER9     g_pIB            = NULL; // Index buffer of the mesh to render
CD3DArcBall                g_ArcBall;               // Arcball for model control
int                        g_iShowUI        = 0;    // 0: show all; 1: hide help; 2: hide captions, 3: hide all
CDXUTDialogResourceManager g_DialogResourceManager; // manager for shared resources of dialogs
CD3DSettingsDlg            g_SettingsDlg;           // Device settings dialog
CDXUTDialog                g_HUD;                   // dialog for standard controls
CDXUTDialog                g_SampleUI;              // dialog for sample specific controls
D3DXVECTOR3                g_vObjectCenter;         // Center of bounding sphere of object
FLOAT                      g_fObjectRadius;         // Radius of bounding sphere of object
D3DXMATRIXA16              g_matView;               // View matrix
D3DXMATRIXA16              g_matProj;               // Projection matrix
float                      g_fZNear;                // Distance of the near clipping plane
float                      g_fZFar;                 // Distance of the far clipping plane
float                      g_fAspect;               // Aspect ratio
DWORD                      g_dwBehaviorFlags;       // Behavior flags of the 3D device
bool                       g_bUseSoftwareVP;        // Flag to indicate whether software vp is
                                                    // required due to lack of hardware
VISMODE                    g_eVisMode = CURVCOLOR;  // Current visualization mode

bool                       g_bUseDiff = true;       // Whether to use diffuse lighting
bool                       g_bWireframe = false;    // Whether to show the mesh in wireframe mode
int                        g_iNumVt;                // Number of vertices of the current model
int                        g_iNumFace;              // number of faces of the current model
int                        g_iVtAttTexSize;         // Size of each dimension of the vertex attribute textures
int                        g_iAdjVtTexSize;         // Size of each dimension of the adjacent vertex texture
LPDIRECT3DTEXTURE9         g_pVtPosTex     = NULL;  // Vertex position texture
LPDIRECT3DTEXTURE9         g_pVtNormalTex  = NULL;  // Vertex normal texture
LPDIRECT3DTEXTURE9         g_pVtTensor2Tex = NULL;  // Tensor of curvature
LPDIRECT3DTEXTURE9         g_pVtTensor3Tex = NULL;  // Tensor of curvature derivative
LPDIRECT3DTEXTURE9         g_pVtP1K1Tex    = NULL;  // #1 principal direction and curvature
LPDIRECT3DTEXTURE9         g_pVtP2K2Tex    = NULL;  // #2 principal direction and curvature
LPDIRECT3DTEXTURE9         g_pAdjVtTex     = NULL;  // 'Adjacent vertices' texture
LPDIRECT3DTEXTURE9         g_pAdjAddTex    = NULL;  // 'Adjacency addresses' texture

LPDIRECT3DSURFACE9         g_pReadSurf     = NULL;  // Temp surface used to read the vertex attributes

LPD3DXEFFECT               g_pGDFx         = NULL;  // D3DX effect interface to the diff. geometry shaders
D3DXHANDLE                 g_hGDAdjVtIdxC  = NULL;  // Handle to 1D-2D constant of vertex adjacency texture
D3DXHANDLE                 g_hGDVtIdxC     = NULL;  // Handle to 1D-2D constant of vertex attributes texture

LPD3DXEFFECT               g_pRenderFx     = NULL;  // D3DX effect interface to the render shaders
D3DXHANDLE                 g_hRenderFxIdxC = NULL;  // Handle to the variable g_fVtIdxC
D3DXHANDLE                 g_hRenderFxIW   = NULL;  // Handle to inverse world matrix
D3DXHANDLE                 g_hRenderFxWVP  = NULL;  // Handle to world*view*proj matrix
D3DXHANDLE                 g_hRenderFxITW  = NULL;  // Handle to inverse transpose of world matrix
D3DXHANDLE                 g_hRenderFxDiff = NULL;  // Handle to the variable g_bUseDiff

TLVERTEX                   g_pScreenQuadVt[4];      // Vertices for the quad rendering on vertex attributes
TLVERTEX                   g_pHelpQuad1   [4];      // Quad to show the caption of the curvature color mapping
TLVERTEX                   g_pHelpQuad2   [4];      // Quad to show the caption of the grayscale mapping
TLVERTEX                   g_pHelpQuad3   [4];      // Quad to show the caption of the curvature derivative color mapping
LPDIRECT3DTEXTURE9         g_pHelpQuad1Tex;         // Texture used with g_pHelpQuad1
LPDIRECT3DTEXTURE9         g_pHelpQuad2Tex;         // Texture used with g_pHelpQuad2
LPDIRECT3DTEXTURE9         g_pHelpQuad3Tex;         // Texture used with g_pHelpQuad3
LPD3DXMESH                 g_pSphere       = NULL;
LPD3DXMESH                 g_pCylinder     = NULL;
LPD3DXMESH                 g_pCylinder2    = NULL;

float                      g_fScale;                // Magic scale value used to determine the size of the 3D widgets

// Pointer to an array of 3D coordinates containing the position
// of each vertex of the original geometry, in object space. The array 
// should be packed like [x1, y1, z1, x2, y2, z2, ..., xn, yn, zn]
float *g_pVtPos = NULL;

// Pointer to an array of 3D coordinates containing the normal vector
// of each vertex of the original geometry, in object space. The array 
// should be packed like [x1, y1, z1, x2, y2, z2, ..., xn, yn, zn]
float *g_pVtNormal = NULL;

// Pointer to an array of 3D coordinates containing the #1 principal
// direction of each vertex, in object space. 
// The array should be packed like [x1, y1, z1, x2, y2, z2, ..., xn, yn, zn]
float *g_pVtP1 = NULL;

// Pointer to an array of triangle face indices. Each set of 3 elements in
// the array contain the indices to a triangle's vertices.
DWORD *g_pFaceIdx = NULL;

////////////////////////////////////////////////////////////////////////////////////////
// UI control IDs
////////////////////////////////////////////////////////////////////////////////////////
#define IDC_TOGGLEFULLSCREEN 1
#define IDC_TOGGLEREF        3
#define IDC_CHANGEDEVICE     4
#define IDC_MESH             5
#define IDC_METHOD           6
#define IDC_DIFFUSE          7
#define IDC_WIREFRAME        8

////////////////////////////////////////////////////////////////////////////////////////
// Forward declarations 
////////////////////////////////////////////////////////////////////////////////////////
bool    CALLBACK IsDeviceAcceptable  ( D3DCAPS9* pCaps, D3DFORMAT AdapterFormat, D3DFORMAT BackBufferFormat, bool bWindowed, void* pUserContext );
bool    CALLBACK ModifyDeviceSettings( DXUTDeviceSettings* pDeviceSettings, const D3DCAPS9* pCaps, void* pUserContext );
HRESULT CALLBACK OnCreateDevice      ( IDirect3DDevice9* pd3dDevice, const D3DSURFACE_DESC* pBackBufferSurfaceDesc, void* pUserContext );
HRESULT CALLBACK OnResetDevice       ( IDirect3DDevice9* pd3dDevice, const D3DSURFACE_DESC* pBackBufferSurfaceDesc, void* pUserContext );
void    CALLBACK OnFrameMove         ( IDirect3DDevice9* pd3dDevice, double fTime, float fElapsedTime, void* pUserContext );
void    CALLBACK OnFrameRender       ( IDirect3DDevice9* pd3dDevice, double fTime, float fElapsedTime, void* pUserContext );
LRESULT CALLBACK MsgProc             ( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, bool* pbNoFurtherProcessing, void* pUserContext );
void    CALLBACK KeyboardProc        ( UINT nChar, bool bKeyDown, bool bAltDown, void* pUserContext );
void    CALLBACK OnGUIEvent          ( UINT nEvent, int nControlID, CDXUTControl* pControl, void* pUserContext );
void    CALLBACK OnLostDevice        ( void* pUserContext );
void    CALLBACK OnDestroyDevice     ( void* pUserContext );

HRESULT CreateRTs     ( );
HRESULT DestroyRTs    ( );
HRESULT CreateNRTs    ( );
HRESULT DestroyNRTs   ( );
HRESULT CreateShaders ( );
HRESULT DestroyShaders( );
HRESULT LoadMesh      ( );
void    ReleaseMesh   ( );

void    InitApp   ( );
HRESULT LoadMesh  ( IDirect3DDevice9* pd3dDevice, WCHAR* strFileName, ID3DXMesh** ppMesh );
void    RenderText( );

////////////////////////////////////////////////////////////////////////////////////////
// Entry point to the program. Initializes everything and goes into a message processing 
// loop. Idle time is used to render the scene.
////////////////////////////////////////////////////////////////////////////////////////
INT WINAPI WinMain( HINSTANCE, HINSTANCE, LPSTR, int )
{
    // Enable run-time memory check for debug builds.
#if defined(DEBUG) | defined(_DEBUG)
    _CrtSetDbgFlag( _CRTDBG_ALLOC_MEM_DF | _CRTDBG_LEAK_CHECK_DF );
#endif

    // Set the callback functions. These functions allow DXUT to notify
    // the application about device changes, user input, and windows messages.  The 
    // callbacks are optional so you need only set callbacks for events you're interested 
    // in. However, if you don't handle the device reset/lost callbacks then the sample 
    // framework won't be able to reset your device since the application must first 
    // release all device resources before resetting.  Likewise, if you don't handle the 
    // device created/destroyed callbacks then DXUT won't be able to 
    // recreate your device resources.
    DXUTSetCallbackDeviceCreated  ( OnCreateDevice  );
    DXUTSetCallbackDeviceReset    ( OnResetDevice   );
    DXUTSetCallbackDeviceLost     ( OnLostDevice    );
    DXUTSetCallbackDeviceDestroyed( OnDestroyDevice );
    DXUTSetCallbackMsgProc        ( MsgProc         );
    DXUTSetCallbackKeyboard       ( KeyboardProc    );
    DXUTSetCallbackFrameRender    ( OnFrameRender   );
    DXUTSetCallbackFrameMove      ( OnFrameMove     );

    // Show the cursor and clip it when in full screen
    DXUTSetCursorSettings( true, true );

    InitApp( );

    // Initialize DXUT and create the desired Win32 window and Direct3D 
    // device for the application. Calling each of these functions is optional, but they
    // allow you to set several options which control the behavior of the framework.
    DXUTInit( true, true, true ); // Parse the command line, handle the default hotkeys, 
								  // and show msgboxes
	DXUTCreateWindow( L"Estimation of differential geometry properties on the GPU" );

    // Supports all types of vertex processing, including mixed.
    DXUTGetEnumeration( )->SetPossibleVertexProcessingList( true, true, true, true );

    DXUTCreateDevice( D3DADAPTER_DEFAULT, true, 640, 480, IsDeviceAcceptable, ModifyDeviceSettings );

    // Pass control to DXUT for handling the message pump and 
    // dispatching render calls. DXUT will call your FrameMove 
    // and FrameRender callback when there is idle time between handling window messages.
    DXUTMainLoop( );

    return DXUTGetExitCode( );
}

////////////////////////////////////////////////////////////////////////////////////////
// Initialize the app 
////////////////////////////////////////////////////////////////////////////////////////
void InitApp( )
{
    // Initialize dialogs
    g_SettingsDlg.Init( &g_DialogResourceManager );
    g_HUD.Init( &g_DialogResourceManager );
    g_SampleUI.Init( &g_DialogResourceManager );

    g_HUD.SetCallback( OnGUIEvent ); int iY = 10; 
    g_HUD.AddButton( IDC_TOGGLEFULLSCREEN, L"Toggle full screen", 35, iY, 125, 22 );
    g_HUD.AddButton( IDC_TOGGLEREF,        L"Toggle REF (F3)",    35, iY += 24, 125, 22 );
    g_HUD.AddButton( IDC_CHANGEDEVICE,     L"Change device (F2)", 35, iY += 24, 125, 22, VK_F2 );

    // Add mixed vp to the available vp choices in device settings dialog.
    DXUTGetEnumeration( )->SetPossibleVertexProcessingList( true, false, false, true );
    
    g_SampleUI.SetCallback( OnGUIEvent ); iY = 10;
    g_SampleUI.AddComboBox( IDC_MESH, 0, iY, 190, 24, L'M' );
    g_SampleUI.GetComboBox( IDC_MESH )->AddItem( L"Teapot",    (void*)TEAPOT    );
    g_SampleUI.GetComboBox( IDC_MESH )->AddItem( L"Knot",      (void*)KNOT      );
    g_SampleUI.GetComboBox( IDC_MESH )->AddItem( L"Head",      (void*)HEAD      );
    g_SampleUI.GetComboBox( IDC_MESH )->AddItem( L"Bunny",     (void*)BUNNY     );
    g_SampleUI.GetComboBox( IDC_MESH )->AddItem( L"Horse",     (void*)HORSE     );
    g_SampleUI.GetComboBox( IDC_MESH )->AddItem( L"Armadillo", (void*)ARMADILLO );
	iY += 25;
    g_SampleUI.AddComboBox( IDC_METHOD, 0, iY, 190, 24, L'V' );
    g_SampleUI.GetComboBox( IDC_METHOD )->AddItem( L"Principal curvatures",   (void*)CURVCOLOR     );
    g_SampleUI.GetComboBox( IDC_METHOD )->AddItem( L"Gaussian curvature",     (void*)GAUSSIANCURV  );
    g_SampleUI.GetComboBox( IDC_METHOD )->AddItem( L"Mean curvature",         (void*)MEANCURV      );
    g_SampleUI.GetComboBox( IDC_METHOD )->AddItem( L"Principal directions",   (void*)PRINCIPALDIRS );
    g_SampleUI.GetComboBox( IDC_METHOD )->AddItem( L"Deriv. curv. magnitude", (void*)MAGCURVDEV    );
    g_SampleUI.GetComboBox( IDC_METHOD )->AddItem( L"None",                   (void*)NONE          );
	iY += 30;
	g_SampleUI.AddCheckBox( IDC_DIFFUSE, L"Diffuse lighting", 5, iY, 245, 15, g_bUseDiff );
	iY += 20;
	g_SampleUI.AddCheckBox( IDC_WIREFRAME, L"Overlaid wireframe", 5, iY, 245, 15, g_bWireframe );
}

////////////////////////////////////////////////////////////////////////////////////////
// Called during device initialization, this code checks the device for some 
// minimum set of capabilities, and rejects those that don't pass by returning false.
////////////////////////////////////////////////////////////////////////////////////////
bool CALLBACK IsDeviceAcceptable( D3DCAPS9* pCaps, D3DFORMAT AdapterFormat, 
                                  D3DFORMAT BackBufferFormat, bool bWindowed, 
								  void* pUserContext )
{
    // Skip backbuffer formats that don't support alpha blending
    IDirect3D9* pD3D = DXUTGetD3DObject( ); 
    if( FAILED( pD3D->CheckDeviceFormat( pCaps->AdapterOrdinal, pCaps->DeviceType,
		AdapterFormat, D3DUSAGE_QUERY_POSTPIXELSHADER_BLENDING, 
        D3DRTYPE_TEXTURE, BackBufferFormat ) ) )
        return false;

    return true;
}

////////////////////////////////////////////////////////////////////////////////////////
// This callback function is called immediately before a device is created to allow the 
// application to modify the device settings. The supplied pDeviceSettings parameter 
// contains the settings that the framework has selected for the new device, and the 
// application can make any desired changes directly to this structure.  Note however 
// that DXUT will not correct invalid device settings so care must be taken 
// to return valid device settings, otherwise IDirect3D9::CreateDevice() will fail.  
////////////////////////////////////////////////////////////////////////////////////////
bool CALLBACK ModifyDeviceSettings( DXUTDeviceSettings* pDeviceSettings, 
								    const D3DCAPS9* pCaps, void* pUserContext )
{
    // If device doesn't support HW T&L or doesn't support 3.0 vertex shaders in HW 
    // then switch to SWVP.
    if( (pCaps->DevCaps & D3DDEVCAPS_HWTRANSFORMANDLIGHT) == 0 ||
         pCaps->VertexShaderVersion < D3DVS_VERSION(3,0) )
    {
        pDeviceSettings->BehaviorFlags = D3DCREATE_SOFTWARE_VERTEXPROCESSING;
    }

    // Debugging vertex shaders requires either REF or software vertex processing 
    // and debugging pixel shaders requires REF.  
#ifdef DEBUG_VS
    if( pDeviceSettings->DeviceType != D3DDEVTYPE_REF )
    {
        pDeviceSettings->BehaviorFlags &= ~D3DCREATE_HARDWARE_VERTEXPROCESSING;
        pDeviceSettings->BehaviorFlags &= ~D3DCREATE_PUREDEVICE;
        pDeviceSettings->BehaviorFlags |= D3DCREATE_SOFTWARE_VERTEXPROCESSING;
    }
#endif
#ifdef DEBUG_PS
    pDeviceSettings->DeviceType = D3DDEVTYPE_REF;
#endif
    // For the first device created if its a REF device, optionally display a warning 
	// dialog box
    static bool s_bFirstTime = true;
    if( s_bFirstTime )
    {
        s_bFirstTime = false;
        if( pDeviceSettings->DeviceType == D3DDEVTYPE_REF )
            DXUTDisplaySwitchingToREFWarning();
    }

    return true;
}

////////////////////////////////////////////////////////////////////////////////////////
// This callback function will be called immediately after the Direct3D device has been 
// created, which will happen during application initialization and windowed/full screen 
// toggles. This is the best location to create D3DPOOL_MANAGED resources since these 
// resources need to be reloaded whenever the device is destroyed. Resources created  
// here should be released in the OnDestroyDevice callback. 
////////////////////////////////////////////////////////////////////////////////////////
HRESULT CALLBACK OnCreateDevice( IDirect3DDevice9* pd3dDevice, 
								 const D3DSURFACE_DESC* pBackBufferSurfaceDesc, 
								 void* pUserContext )
{
    HRESULT hr;

	g_pd3dDevice = pd3dDevice;

    V_RETURN( g_DialogResourceManager.OnCreateDevice( pd3dDevice ) );
    V_RETURN( g_SettingsDlg.OnCreateDevice( pd3dDevice ) );

    // Initialize the font
    V_RETURN( D3DXCreateFont( pd3dDevice, 15, 0, FW_BOLD, 0, FALSE, DEFAULT_CHARSET, 
							  OUT_DEFAULT_PRECIS, DEFAULT_QUALITY, 
							  DEFAULT_PITCH | FF_DONTCARE, L"Arial", &g_pFont ) );

	V_RETURN( LoadMesh( ) );

    // Obtain the behavior flags
    D3DDEVICE_CREATION_PARAMETERS cp;
    pd3dDevice->GetCreationParameters( &cp );
    g_dwBehaviorFlags = cp.BehaviorFlags;

	// Load the captions
	WCHAR str[MAX_PATH];
	V_RETURN( DXUTFindDXSDKMediaFileCch( str, MAX_PATH, L"caption1.dds" ) );
	V_RETURN( D3DXCreateTextureFromFile( g_pd3dDevice, str, &g_pHelpQuad1Tex ) );

	V_RETURN( DXUTFindDXSDKMediaFileCch( str, MAX_PATH, L"caption2.dds" ) );
	V_RETURN( D3DXCreateTextureFromFile( g_pd3dDevice, str, &g_pHelpQuad2Tex ) );

	V_RETURN( DXUTFindDXSDKMediaFileCch( str, MAX_PATH, L"caption3.dds" ) );
	V_RETURN( D3DXCreateTextureFromFile( g_pd3dDevice, str, &g_pHelpQuad3Tex ) );

    return S_OK;
}

////////////////////////////////////////////////////////////////////////////////////////
// Load the currently selected mesh. 
////////////////////////////////////////////////////////////////////////////////////////
HRESULT LoadMesh( )
{
    HRESULT hr;

	// Load the mesh
	WCHAR str[MAX_PATH];
	V_RETURN( DXUTFindDXSDKMediaFileCch( str, MAX_PATH, g_szMesh[ g_eCurMesh ] ) );
	WCHAR strPath[MAX_PATH];
	StringCchCopy( strPath, MAX_PATH, str );
	WCHAR* pLastSlash = wcsrchr( strPath, L'\\' );
	if( pLastSlash )
	{
		*pLastSlash = 0;
		++pLastSlash;
	}
	else
	{
		StringCchCopy( strPath, MAX_PATH, L"." );
		pLastSlash = str;
	}
	WCHAR strCWD[MAX_PATH];
	GetCurrentDirectory( MAX_PATH, strCWD );
	SetCurrentDirectory( strPath );
	V_RETURN( D3DXLoadMeshFromX( pLastSlash, D3DXMESH_MANAGED, g_pd3dDevice,
		NULL, NULL, NULL, NULL, &g_pMesh ) );

	g_iNumVt   = g_pMesh->GetNumVertices( );
	g_iNumFace = g_pMesh->GetNumFaces   ( );

	// Set up the vertex ID in the vertex buffer
	LPD3DXMESH pCloneMesh;
	V_RETURN( g_pMesh->CloneMeshFVF( D3DXMESH_32BIT | D3DXMESH_MANAGED, 
		g_pMesh->GetFVF( ) | FVF_MESHVERTEX, 
		g_pd3dDevice, &pCloneMesh ) );

	MESHVERTEX *pData;
	V_RETURN( pCloneMesh->LockVertexBuffer( 0, (LPVOID*) &pData ) );

	for( int i = 0; i < g_iNumVt; i++ ) 
		pData[i].id = (float) i;

	// Keep a copy of the vertex buffer and index buffer
	pCloneMesh->GetVertexBuffer( &g_pVB );
	pCloneMesh->GetIndexBuffer ( &g_pIB );

	// Compute bounding sphere
	V_RETURN( D3DXComputeBoundingSphere( &pData[0].p, g_iNumVt,
		pCloneMesh->GetNumBytesPerVertex( ), &g_vObjectCenter, &g_fObjectRadius ) );

	// Compute minimum size required by a vertex attribute texture
	DWORD dwPow = 1;
	g_iVtAttTexSize = 1;
	while( 1 )
	{
		g_iVtAttTexSize = 1 << dwPow++;
		if( g_iVtAttTexSize * g_iVtAttTexSize >= g_iNumVt )
			break;
	}

	// Allocate temporary arrays for the vertex textures
	g_pVtPos    = new float[ g_iNumVt   * 3 ];
	g_pVtNormal = new float[ g_iNumVt   * 3 ];
	g_pVtP1     = new float[ g_iNumVt   * 3 ];
	g_pFaceIdx  = new DWORD[ g_iNumFace * 3 ];
	if( !g_pVtPos || !g_pVtNormal || !g_pFaceIdx )
	{
		SAFE_DELETE_ARRAY( g_pVtPos );
		SAFE_DELETE_ARRAY( g_pVtNormal );
		SAFE_DELETE_ARRAY( g_pFaceIdx );

		V_RETURN( pCloneMesh->UnlockVertexBuffer( ) );

		return E_OUTOFMEMORY;
	}

	memset( g_pVtPos,    0, sizeof( float ) * g_iNumVt   * 3 );
	memset( g_pVtNormal, 0, sizeof( float ) * g_iNumVt   * 3 );
	memset( g_pVtP1,     0, sizeof( float ) * g_iNumVt   * 3 );
	memset( g_pFaceIdx,  0, sizeof( DWORD ) * g_iNumFace * 3 );

	// Fill g_pVtPos
	int iOffset = 0;
	for( int i = 0; i < g_iNumVt; i++ )
	{
		g_pVtPos[iOffset++] = pData[i].p.x;
		g_pVtPos[iOffset++] = pData[i].p.y;
		g_pVtPos[iOffset++] = pData[i].p.z;
	}

	// Fill g_pVtNormal
	iOffset = 0;
	for( int i = 0; i < g_iNumVt; i++ )
	{
		g_pVtNormal[iOffset++] = pData[i].n.x;
		g_pVtNormal[iOffset++] = pData[i].n.y;
		g_pVtNormal[iOffset++] = pData[i].n.z;
	}

	V_RETURN( pCloneMesh->UnlockVertexBuffer( ) );

	DWORD *pIdxData;
	V_RETURN( pCloneMesh->LockIndexBuffer( D3DLOCK_READONLY, (LPVOID*) &pIdxData ) );

	// Fill g_pFaceIdx
	memcpy( g_pFaceIdx, pIdxData, sizeof( DWORD ) * g_iNumFace * 3 );

	V_RETURN( pCloneMesh->UnlockIndexBuffer( ) );

	SAFE_RELEASE( g_pMesh );
	g_pMesh = pCloneMesh;

	// Create sphere
	g_fScale = g_fObjectRadius / 180.0f;
	V_RETURN( D3DXCreateSphere( g_pd3dDevice, g_fScale, 10, 10, &g_pSphere, NULL ) );
	// Create cylinder
	V_RETURN( D3DXCreateCylinder( g_pd3dDevice, g_fScale / 1.5f, g_fScale / 1.5f, 
		g_fScale * 6.0f, 10, 1, &g_pCylinder, NULL ) );
	// Create arrow
	V_RETURN( D3DXCreateCylinder( g_pd3dDevice, 0.0f, g_fScale, 
		g_fScale * 2.0f, 10, 1, &g_pCylinder2, NULL ) );

	SetCurrentDirectory( strCWD );

    return S_OK;
}

////////////////////////////////////////////////////////////////////////////////////////
// Release the currently loaded mesh. 
////////////////////////////////////////////////////////////////////////////////////////
void ReleaseMesh( )
{
	SAFE_RELEASE( g_pCylinder2 );
	SAFE_RELEASE( g_pCylinder  );
	SAFE_RELEASE( g_pSphere    );

	SAFE_RELEASE( g_pVB   );
	SAFE_RELEASE( g_pIB   );
	SAFE_RELEASE( g_pMesh );

	SAFE_DELETE_ARRAY( g_pVtPos    );
	SAFE_DELETE_ARRAY( g_pVtNormal );
	SAFE_DELETE_ARRAY( g_pVtP1     );
	SAFE_DELETE_ARRAY( g_pFaceIdx  );
}

////////////////////////////////////////////////////////////////////////////////////////
// This callback function will be called immediately after the Direct3D device has been 
// reset, which will happen after a lost device scenario. This is the best location to 
// create D3DPOOL_DEFAULT resources since these resources need to be reloaded whenever 
// the device is lost. Resources created here should be released in the OnLostDevice 
// callback. 
////////////////////////////////////////////////////////////////////////////////////////
HRESULT CALLBACK OnResetDevice( IDirect3DDevice9* pd3dDevice, 
                                const D3DSURFACE_DESC* pBackBufferSurfaceDesc, 
								void* pUserContext )
{
    HRESULT hr;

    V_RETURN( g_DialogResourceManager.OnResetDevice( ) );
    V_RETURN( g_SettingsDlg.OnResetDevice( ) );

    if( g_pFont )
        V_RETURN( g_pFont->OnResetDevice( ) );
    if( g_pEffect )
        V_RETURN( g_pEffect->OnResetDevice( ) );

    // Create a sprite to help batch calls when drawing many lines of text
    V_RETURN( D3DXCreateSprite( pd3dDevice, &g_pTextSprite ) );

    // Setup render state
    pd3dDevice->SetRenderState ( D3DRS_LIGHTING,         TRUE           );
    pd3dDevice->SetRenderState ( D3DRS_DITHERENABLE,     TRUE           );
    pd3dDevice->SetRenderState ( D3DRS_ZENABLE,          TRUE           );
    pd3dDevice->SetRenderState ( D3DRS_CULLMODE,         D3DCULL_NONE   );
    pd3dDevice->SetRenderState ( D3DRS_AMBIENT,          0x33333333     );
    pd3dDevice->SetRenderState ( D3DRS_NORMALIZENORMALS, TRUE           );
    pd3dDevice->SetSamplerState( 0, D3DSAMP_MAGFILTER,   D3DTEXF_LINEAR );
    pd3dDevice->SetSamplerState( 0, D3DSAMP_MINFILTER,   D3DTEXF_LINEAR );

    // Setup the projection matrix
    g_fAspect = (float)pBackBufferSurfaceDesc->Width / 
				(float)pBackBufferSurfaceDesc->Height;
	g_fZNear = g_fObjectRadius / 64.0f;
	g_fZFar  = g_fObjectRadius * 200.0f;
    D3DXMatrixPerspectiveFovLH( &g_matProj, D3DX_PI / 4, g_fAspect,
                                g_fZNear, g_fZFar );
    pd3dDevice->SetTransform( D3DTS_PROJECTION, &g_matProj );

    // Setup the arcball parameters
    g_ArcBall.SetWindow( pBackBufferSurfaceDesc->Width, 
						 pBackBufferSurfaceDesc->Height, 0.85f );
    g_ArcBall.SetTranslationRadius( g_fObjectRadius );

    g_HUD.SetLocation( pBackBufferSurfaceDesc->Width-170, 0 );
    g_HUD.SetSize( 170, 170 );
    g_SampleUI.SetLocation( 3, 45 );
    g_SampleUI.SetSize( 240, 70 );

	// Set up UI screen space quads used to show captions
	int x, y;

	x = pBackBufferSurfaceDesc->Width  - 138;
	y = pBackBufferSurfaceDesc->Height - 132;
	g_pHelpQuad1[0].p = D3DXVECTOR4( x +   0.0f - .5f, y +  0.0f - .5f, 0.0f, 1.0f );
	g_pHelpQuad1[1].p = D3DXVECTOR4( x + 128.0f - .5f, y +  0.0f - .5f, 0.0f, 1.0f );
	g_pHelpQuad1[2].p = D3DXVECTOR4( x +   0.0f - .5f, y + 128.0f - .5f, 0.0f, 1.0f );
	g_pHelpQuad1[3].p = D3DXVECTOR4( x + 128.0f - .5f, y + 128.0f - .5f, 0.0f, 1.0f );

	g_pHelpQuad1[0].t = D3DXVECTOR2( 0.0f, 0.0f );
	g_pHelpQuad1[1].t = D3DXVECTOR2( 1.0f, 0.0f );
	g_pHelpQuad1[2].t = D3DXVECTOR2( 0.0f, 1.0f );
	g_pHelpQuad1[3].t = D3DXVECTOR2( 1.0f, 1.0f );

	x = pBackBufferSurfaceDesc->Width  - 138;
	y = pBackBufferSurfaceDesc->Height - 69;
	g_pHelpQuad2[0].p = D3DXVECTOR4( x +   0.0f - .5f, y +  0.0f - .5f, 0.0f, 1.0f );
	g_pHelpQuad2[1].p = D3DXVECTOR4( x + 128.0f - .5f, y +  0.0f - .5f, 0.0f, 1.0f );
	g_pHelpQuad2[2].p = D3DXVECTOR4( x +   0.0f - .5f, y + 64.0f - .5f, 0.0f, 1.0f );
	g_pHelpQuad2[3].p = D3DXVECTOR4( x + 128.0f - .5f, y + 64.0f - .5f, 0.0f, 1.0f );

	g_pHelpQuad2[0].t = D3DXVECTOR2( 0.0f, 0.0f );
	g_pHelpQuad2[1].t = D3DXVECTOR2( 1.0f, 0.0f );
	g_pHelpQuad2[2].t = D3DXVECTOR2( 0.0f, 1.0f );
	g_pHelpQuad2[3].t = D3DXVECTOR2( 1.0f, 1.0f );

	x = pBackBufferSurfaceDesc->Width  - 135;
	y = pBackBufferSurfaceDesc->Height - 50;
	g_pHelpQuad3[0].p = D3DXVECTOR4( x +   0.0f - .5f, y +  0.0f - .5f, 0.0f, 1.0f );
	g_pHelpQuad3[1].p = D3DXVECTOR4( x + 128.0f - .5f, y +  0.0f - .5f, 0.0f, 1.0f );
	g_pHelpQuad3[2].p = D3DXVECTOR4( x +   0.0f - .5f, y + 32.0f - .5f, 0.0f, 1.0f );
	g_pHelpQuad3[3].p = D3DXVECTOR4( x + 128.0f - .5f, y + 32.0f - .5f, 0.0f, 1.0f );

	g_pHelpQuad3[0].t = D3DXVECTOR2( 0.0f, 0.0f );
	g_pHelpQuad3[1].t = D3DXVECTOR2( 1.0f, 0.0f );
	g_pHelpQuad3[2].t = D3DXVECTOR2( 0.0f, 1.0f );
	g_pHelpQuad3[3].t = D3DXVECTOR2( 1.0f, 1.0f );

	V_RETURN( CreateRTs    ( ) );
	V_RETURN( CreateNRTs   ( ) );
	V_RETURN( CreateShaders( ) );

    return S_OK;
}

////////////////////////////////////////////////////////////////////////////////////////
// This callback function will be called once at the beginning of every frame. This is 
// the best location for your application to handle updates to the scene, but is not 
// intended to contain actual rendering calls, which should instead be placed in the 
// OnFrameRender callback.  
////////////////////////////////////////////////////////////////////////////////////////
void CALLBACK OnFrameMove( IDirect3DDevice9* pd3dDevice, double fTime, 
						   float fElapsedTime, void* pUserContext )
{
     // Setup world matrix
    D3DXMATRIXA16 matWorld;
    D3DXMatrixTranslation( &matWorld, -g_vObjectCenter.x,
                                      -g_vObjectCenter.y,
                                      -g_vObjectCenter.z );
    D3DXMatrixMultiply( &matWorld, &matWorld, g_ArcBall.GetRotationMatrix( ) );
    D3DXMatrixMultiply( &matWorld, &matWorld, g_ArcBall.GetTranslationMatrix( ) );
    pd3dDevice->SetTransform( D3DTS_WORLD, &matWorld );

    D3DXVECTOR3 vEye( 0, 50, -2*g_fObjectRadius );
    D3DXVECTOR3 vAt ( 0, 0, 0 );
    D3DXVECTOR3 vUp ( 0, 1, 0 );
    D3DXMatrixLookAtLH( &g_matView, &vEye, &vAt, &vUp);

    pd3dDevice->SetTransform( D3DTS_VIEW,  &g_matView );
}

////////////////////////////////////////////////////////////////////////////////////////
// This callback function will be called at the end of every frame to perform all the 
// rendering calls for the scene, and it will also be called if the window needs to be 
// repainted. After this function has returned, DXUT will call 
// IDirect3DDevice9::Present to display the contents of the next buffer in the swap chain
////////////////////////////////////////////////////////////////////////////////////////
void CALLBACK OnFrameRender( IDirect3DDevice9* pd3dDevice, double fTime, 
							 float fElapsedTime, void* pUserContext )
{
    // If the settings dialog is being shown, then
    // render it instead of rendering the app's scene
    if( g_SettingsDlg.IsActive( ) )
    {
        g_SettingsDlg.OnRender( fElapsedTime );
        return;
    }

    HRESULT hr;

	//
	// Compute differential geometry properties
	//

	// Disable z-buffering (testing and writing) 
	g_pd3dDevice->SetRenderState( D3DRS_ZENABLE,      FALSE );
	g_pd3dDevice->SetRenderState( D3DRS_ZWRITEENABLE, FALSE );

	// Set viewport for 1:1 pixel=texel=geometry mapping
	D3DVIEWPORT9 d3dPrevVp, d3dCurVp;
	g_pd3dDevice->GetViewport( &d3dPrevVp );
	d3dCurVp = d3dPrevVp;
	d3dCurVp.X = d3dCurVp.Y = 0;
	d3dCurVp.Width = d3dCurVp.Height = g_iVtAttTexSize;
	g_pd3dDevice->SetViewport( &d3dCurVp );

	// Compute 2nd order properties (tensor of curvature, principal directions
	// and curvatures
	LPDIRECT3DSURFACE9 pd3dVtCacheSur[4] = {NULL};

	// Render target 0: 2nd fundamental form
	g_pVtTensor2Tex->GetSurfaceLevel( 0, &pd3dVtCacheSur[0] );
	g_pd3dDevice->SetRenderTarget( 0, pd3dVtCacheSur[0] );

	// Render target 1: #1 principle direction and curvature
	g_pVtP1K1Tex->GetSurfaceLevel( 0, &pd3dVtCacheSur[1] );
	g_pd3dDevice->SetRenderTarget( 1, pd3dVtCacheSur[1] );

	// Render target 2: #2 principle direction and curvature
	g_pVtP2K2Tex->GetSurfaceLevel( 0, &pd3dVtCacheSur[2] );
	g_pd3dDevice->SetRenderTarget( 2, pd3dVtCacheSur[2] );

	// Texture 0: vertex position
	g_pGDFx->SetTexture( "Tex0", g_pVtPosTex    );
	// Texture 1: normal vertex
	g_pGDFx->SetTexture( "Tex1", g_pVtNormalTex );
	// Texture 2: adjacency addresses
	g_pGDFx->SetTexture( "Tex2", g_pAdjAddTex   );
	// Texture 3: vertex adjacency addresses
	g_pGDFx->SetTexture( "Tex3", g_pAdjVtTex    );

	g_pd3dDevice->SetFVF( FVF_TLVERTEX );
	g_pd3dDevice->SetStreamSource( 0, NULL, 0, sizeof( TLVERTEX ) );

	// Set uniform parameters
	float pFloat[2];

	// 1D-2D constant of vertex attributes texture
	pFloat[0] = 1.0f / (float) g_iVtAttTexSize;
	pFloat[1] = 1.0f / (float)( g_iVtAttTexSize * g_iVtAttTexSize );
	g_pGDFx->SetFloatArray( g_hGDVtIdxC, pFloat, 2 );

	// 1D-2D constant of vertex adjacency texture
	pFloat[0] = 1.0f / (float) g_iAdjVtTexSize; 
	pFloat[1] = 1.0f / (float)( g_iAdjVtTexSize * g_iAdjVtTexSize );
	g_pGDFx->SetFloatArray( g_hGDAdjVtIdxC, pFloat, 2 );

	// Set effect
	V( g_pGDFx->SetTechnique( "Tensor2" ) );

	UINT uiPasses;
	V( g_pGDFx->Begin( &uiPasses, 0 ) );

	for( UINT iPass = 0; iPass < uiPasses; iPass++ )
	{
		V( g_pGDFx->BeginPass( iPass ) );

		// Render filled quad to hit every pixel/texel
		g_pd3dDevice->DrawPrimitiveUP( D3DPT_TRIANGLESTRIP, 2, 
			g_pScreenQuadVt, sizeof( TLVERTEX ) );

		g_pGDFx->EndPass( );
	}
	g_pGDFx->End( );

	SAFE_RELEASE( pd3dVtCacheSur[0] );
	SAFE_RELEASE( pd3dVtCacheSur[1] );
	SAFE_RELEASE( pd3dVtCacheSur[2] );

	if( g_eVisMode == MAGCURVDEV )
	{
		// Compute 3rd order properties (tensor of derivative of curvature)

		// Render target 0: tensor of derivative of curvature
		g_pVtTensor3Tex->GetSurfaceLevel( 0, &pd3dVtCacheSur[3] );
		g_pd3dDevice->SetRenderTarget( 0, pd3dVtCacheSur[3] );

		// Texture 0: vertex position
		g_pGDFx->SetTexture( "Tex0", g_pVtPosTex    );
		// Texture 1: normal vertex
		g_pGDFx->SetTexture( "Tex1", g_pVtNormalTex );
		// Texture 2: adjacency addresses
		g_pGDFx->SetTexture( "Tex2", g_pAdjAddTex   );
		// Texture 3: vertex adjacency addresses
		g_pGDFx->SetTexture( "Tex3", g_pAdjVtTex    );
		// Texture 4: Principal direction and curvature #1
		g_pGDFx->SetTexture( "Tex4", g_pVtP1K1Tex   );
		// Texture 5: Principal direction and curvature #2
		g_pGDFx->SetTexture( "Tex5", g_pVtP2K2Tex   );

		g_pd3dDevice->SetFVF( FVF_TLVERTEX );
		g_pd3dDevice->SetStreamSource( 0, NULL, 0, sizeof( TLVERTEX ) );

		// Set uniform parameters
		float pFloat[2];

		// 1D-2D constant of vertex attributes texture
		pFloat[0] = 1.0f / (float) g_iVtAttTexSize;
		pFloat[1] = 1.0f / (float)( g_iVtAttTexSize * g_iVtAttTexSize );
		g_pGDFx->SetFloatArray( g_hGDVtIdxC, pFloat, 2 );

		// 1D-2D constant of vertex adjacency texture
		pFloat[0] = 1.0f / (float) g_iAdjVtTexSize; 
		pFloat[1] = 1.0f / (float)( g_iAdjVtTexSize * g_iAdjVtTexSize );
		g_pGDFx->SetFloatArray( g_hGDAdjVtIdxC, pFloat, 2 );

		// Set effect
		V( g_pGDFx->SetTechnique( "Tensor3" ) );

		UINT uiPasses;
		V( g_pGDFx->Begin( &uiPasses, 0 ) );

		for( UINT iPass = 0; iPass < uiPasses; iPass++ )
		{
			V( g_pGDFx->BeginPass( iPass ) );

			// Render filled quad to hit every pixel/texel
			g_pd3dDevice->DrawPrimitiveUP( D3DPT_TRIANGLESTRIP, 2, 
				g_pScreenQuadVt, sizeof( TLVERTEX ) );

			g_pGDFx->EndPass( );
		}
		g_pGDFx->End( );

		SAFE_RELEASE( pd3dVtCacheSur[3] );
	}

	// Re-enable z-buffering (testing and writing)
	g_pd3dDevice->SetRenderState( D3DRS_ZWRITEENABLE, TRUE );
	g_pd3dDevice->SetRenderState( D3DRS_ZENABLE,      TRUE );

	// Restore viewport
	g_pd3dDevice->SetViewport( &d3dPrevVp );

	// Fill g_pVtP1
	LPDIRECT3DSURFACE9 pRTSurf; 
	V( g_pVtP1K1Tex->GetSurfaceLevel( 0, &pRTSurf ) );
	V( g_pd3dDevice->GetRenderTargetData( pRTSurf, g_pReadSurf ) );

	D3DLOCKED_RECT d3dLockedRect;
	V( g_pReadSurf->LockRect( &d3dLockedRect, NULL, D3DLOCK_READONLY ) );

	int iCurVt = 0, iOffset = 0;

	for( int iY = 0; iY < g_iVtAttTexSize; iY++ )
	{
		for( int iX = 0; iX < g_iVtAttTexSize; iX++ )
		{
			if( iCurVt == g_iNumVt )
				break;

			BYTE *pLockedRectB = (BYTE*) d3dLockedRect.pBits;
			float *vTexel = (float*) ( pLockedRectB + 
				(iX*4*sizeof(float) + iY*d3dLockedRect.Pitch ) );

			g_pVtP1[iOffset++] = vTexel[0];
			g_pVtP1[iOffset++] = vTexel[1];
			g_pVtP1[iOffset++] = vTexel[2];

			iCurVt++;
		}
	}

	V( g_pReadSurf->UnlockRect( ) );

	SAFE_RELEASE( pRTSurf );


	//
	// Render actual geometry onto the backbuffer
	//

	g_pd3dDevice->SetRenderTarget( 0, g_pd3dBackBuffer );
	g_pd3dDevice->SetRenderTarget( 1, NULL );
	g_pd3dDevice->SetRenderTarget( 2, NULL );

	// Get current world, view, proj matrices
	D3DXMATRIX mWorld, mITW, mIW;
	g_pd3dDevice->GetTransform( D3DTS_WORLD, &mWorld );
	D3DXMATRIX mWVP = mWorld * g_matView * g_matProj;

	float fDet;
	D3DXMatrixInverse  ( &mIW, &fDet, &mWorld );
	D3DXMatrixTranspose( &mITW, &mIW );

	// 1D-2D index conversion constant for sampling vertex 
	// attributes textures
	pFloat[0] = 1.0f / (float) g_iVtAttTexSize;
	pFloat[1] = 1.0f / (float)( g_iVtAttTexSize * g_iVtAttTexSize );

    // Clear the backbuffer
    V( pd3dDevice->Clear( 0L, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER,
                          D3DCOLOR_ARGB(255, 32, 32, 32), 1.0f, 0L ) );

    // Begin the scene
    if( SUCCEEDED( pd3dDevice->BeginScene( ) ) )
    {
		// Set 1D-2D index conversion constant
		g_pRenderFx->SetFloatArray( g_hRenderFxIdxC, pFloat, 2 );

		// Set transformation matrices to the current shader
		g_pRenderFx->SetMatrix( g_hRenderFxIW,   &mIW       );
		g_pRenderFx->SetMatrix( g_hRenderFxWVP,  &mWVP      );
		g_pRenderFx->SetMatrix( g_hRenderFxITW,  &mITW      );
		g_pRenderFx->SetBool  ( g_hRenderFxDiff, g_bUseDiff );

		// Texture 0: vertex position
		g_pRenderFx->SetTexture( "Tex0", g_pVtPosTex );
		// Texture 1: vertex normal
		g_pRenderFx->SetTexture( "Tex1", g_pVtNormalTex );

		// Set the texture samplers according to the 
		// properties selected to be visualized
		switch( g_eVisMode )
		{
		case CURVCOLOR:
		case GAUSSIANCURV:
		case MEANCURV:
		case PRINCIPALDIRS:
			g_pRenderFx->SetTexture( "Tex2", g_pVtP1K1Tex );
			g_pRenderFx->SetTexture( "Tex3", g_pVtP2K2Tex );
			break;
		case MAGCURVDEV:
			g_pRenderFx->SetTexture( "Tex4", g_pVtTensor3Tex );
			break;
		}

		// Set the effect according to the properties selected 
		// to be visualized
		switch( g_eVisMode )
		{
		case CURVCOLOR:
			V( g_pRenderFx->SetTechnique( "Render0" ) );
			break;
		case GAUSSIANCURV:
			V( g_pRenderFx->SetTechnique( "Render2" ) );
			break;
		case MEANCURV:
			V( g_pRenderFx->SetTechnique( "Render3" ) );
			break;
		case MAGCURVDEV:
			V( g_pRenderFx->SetTechnique( "Render4" ) );
			break;
		case PRINCIPALDIRS:
		default:
			V( g_pRenderFx->SetTechnique( "Render5" ) );
			break;
		}

		UINT uiPasses;
		V( g_pRenderFx->Begin( &uiPasses, 0 ) );

		for( UINT iPass = 0; iPass < uiPasses; iPass++ )
		{
			V( g_pRenderFx->BeginPass( iPass ) );

			g_pd3dDevice->SetFVF( FVF_MESHVERTEX );
			g_pd3dDevice->SetStreamSource( 0, g_pVB, 0, sizeof( MESHVERTEX ) );
			g_pd3dDevice->SetIndices( g_pIB );

			// Draw mesh
			V( g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, 0, 
				g_iNumVt, 0, g_iNumFace ) );

			g_pRenderFx->EndPass( );
		}

		g_pRenderFx->End( );

		if( g_bWireframe )
		{
			g_pd3dDevice->SetRenderState( D3DRS_LIGHTING, FALSE );
			g_pd3dDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_WIREFRAME );

			D3DXMatrixPerspectiveFovLH( &g_matProj, D3DX_PI / 4, g_fAspect,
										g_fZNear + 0.001f, g_fZFar + 0.001f );
			g_pd3dDevice->SetTransform( D3DTS_PROJECTION, &g_matProj );

			// Draw mesh
			V( g_pd3dDevice->DrawIndexedPrimitive( D3DPT_TRIANGLELIST, 0, 0, 
				g_iNumVt, 0, g_iNumFace ) );

			D3DXMatrixPerspectiveFovLH( &g_matProj, D3DX_PI / 4, g_fAspect,
										g_fZNear - 0.001f, g_fZFar - 0.001f );
			g_pd3dDevice->SetTransform( D3DTS_PROJECTION, &g_matProj );

			g_pd3dDevice->SetRenderState( D3DRS_FILLMODE, D3DFILL_SOLID );
			g_pd3dDevice->SetRenderState( D3DRS_LIGHTING, TRUE );
		}

		if( g_eVisMode == PRINCIPALDIRS )
		{
			// Create light source
			D3DCOLORVALUE d3dLightDif = { 1.0f, 1.0f, 1.0f, 1.0f };
			D3DCOLORVALUE d3dLightAmb = { 0.05f, 0.05f, 0.05f, 1.0f };
			D3DVECTOR d3dLightDir = { -1.0f, -1.0f, 1.0f };

			D3DLIGHT9 d3dLight;
			d3dLight.Type      = D3DLIGHT_DIRECTIONAL;
			d3dLight.Diffuse   = d3dLightDif;
			d3dLight.Ambient   = d3dLightAmb;
			d3dLight.Direction = d3dLightDir;

			// Create material
			D3DMATERIAL9 d3dMat;
			D3DCOLORVALUE d3dMatColor = { 1.0f, 1.0f, 1.0f, 1.0f };
			ZeroMemory( &d3dMat, sizeof( d3dMat ) );
			d3dMat.Diffuse = d3dMat.Ambient = d3dMatColor;
			V( g_pd3dDevice->SetMaterial( &d3dMat ) );

			V( g_pd3dDevice->SetLight( 0, &d3dLight ) );
			V( g_pd3dDevice->LightEnable( 0, TRUE ) );
			V( g_pd3dDevice->SetRenderState( D3DRS_LIGHTING, TRUE ) );

			// Render tangent basis aligned to principal directions
			for( int i = 0; i < g_iNumVt; i++ )
			{
				D3DXMATRIX mW;
				D3DXMatrixTranslation( &mW, g_pVtPos[i*3+0], g_pVtPos[i*3+1], g_pVtPos[i*3+2] );
				D3DXMatrixMultiply( &mW, &mW, &mWorld );

				//
				D3DXMATRIX matTemp;
				D3DXVECTOR3 m_vTriadTangent = D3DXVECTOR3( 
					g_pVtP1[i*3+0], g_pVtP1[i*3+1], g_pVtP1[i*3+2] );
				D3DXVECTOR3 m_vTriadNormal = D3DXVECTOR3( 
					g_pVtNormal[i*3+0], g_pVtNormal[i*3+1], g_pVtNormal[i*3+2] );

				// Align triad
				D3DXVECTOR3 vDir = m_vTriadNormal;
				D3DXVECTOR3 vXAxis, vYAxis, vZAxis;
				D3DXVECTOR3 vUp = m_vTriadTangent;

				// Orthogonalize the frame 
				D3DXVECTOR3 vTmp = vUp - m_vTriadNormal * D3DXVec3Dot( &m_vTriadNormal, &vUp );
				D3DXVec3Normalize( &vUp, &vTmp );

				vYAxis = vDir;
				if( vDir == vUp )
					vZAxis = D3DXVECTOR3( 0.0f, 0.0f, 1.0f );
				else
					D3DXVec3Cross( &vZAxis, &vDir, &vUp );

				D3DXVec3Normalize( &vZAxis, &vZAxis );
				D3DXVec3Cross    ( &vXAxis, &vZAxis, &vYAxis );
				D3DXVec3Normalize( &vXAxis, &vXAxis );

				D3DXMatrixIdentity( &matTemp );
				matTemp._11 = vXAxis.x; matTemp._12 = vXAxis.y;	matTemp._13 = vXAxis.z;
				matTemp._21 = vYAxis.x;	matTemp._22 = vYAxis.y;	matTemp._23 = vYAxis.z;
				matTemp._31 = vZAxis.x;	matTemp._32 = vZAxis.y;	matTemp._33 = vZAxis.z;
				D3DXMatrixMultiply( &mW, &matTemp, &mW );

				g_pd3dDevice->SetTransform( D3DTS_WORLD, &mW );

				// Sphere centered at the vertex
				d3dLight.Diffuse.r = 1.0f;
				d3dLight.Diffuse.g = 1.0f;
				d3dLight.Diffuse.b = 1.0f;
				V( g_pd3dDevice->SetLight( 0, &d3dLight ) );

				g_pSphere->DrawSubset( 0 );

				//

				D3DXMATRIX matRes;

				// Minimum direction
				d3dLight.Diffuse.r = 1.0f;
				d3dLight.Diffuse.g = 0.0f;
				d3dLight.Diffuse.b = 0.0f;
				V( g_pd3dDevice->SetLight( 0, &d3dLight ) );

				D3DXMatrixTranslation( &matTemp, 0.0f, 0.0f, -(g_fScale * 3.0f) );
				D3DXMatrixMultiply( &matRes, &matTemp, &mW );
				g_pd3dDevice->SetTransform( D3DTS_WORLD, &matRes );
				g_pCylinder->DrawSubset( 0 );

				D3DXMatrixTranslation( &matTemp, 0.0f, 0.0f, -(g_fScale * 7.0f) );
				D3DXMatrixMultiply( &matRes, &matTemp, &mW );
				g_pd3dDevice->SetTransform( D3DTS_WORLD, &matRes );
				g_pCylinder2->DrawSubset( 0 );

				// Maximum direction
				d3dLight.Diffuse.r = 0.0f;
				d3dLight.Diffuse.g = 1.0f;
				d3dLight.Diffuse.b = 0.0f;
				V( g_pd3dDevice->SetLight( 0, &d3dLight ) );

				D3DXMatrixRotationY( &matTemp, -D3DX_PI / 2.0f );
				D3DXMatrixMultiply( &matRes, &matTemp, &mW );
				D3DXMatrixTranslation( &matTemp, 0.0f, 0.0f, -(g_fScale * 3.0f) );
				D3DXMatrixMultiply( &matRes, &matTemp, &matRes );
				g_pd3dDevice->SetTransform( D3DTS_WORLD, &matRes );
				g_pCylinder->DrawSubset( 0 );

				D3DXMatrixRotationY( &matTemp, -D3DX_PI / 2.0f );
				D3DXMatrixMultiply( &matRes, &matTemp, &mW );
				D3DXMatrixTranslation( &matTemp, 0.0f, 0.0f, -(g_fScale * 7.0f) );
				D3DXMatrixMultiply( &matRes, &matTemp, &matRes );
				g_pd3dDevice->SetTransform( D3DTS_WORLD, &matRes );
				g_pCylinder2->DrawSubset( 0 );
			}
		}

        RenderText( );
        V( g_HUD.OnRender( fElapsedTime ) );
        V( g_SampleUI.OnRender( fElapsedTime ) );

		if( g_iShowUI < 3 )
		{
			switch( g_eVisMode )
			{
			case CURVCOLOR:
				g_pd3dDevice->SetFVF( FVF_TLVERTEX );
				g_pd3dDevice->SetStreamSource( 0, NULL, 0, sizeof( TLVERTEX ) );
				g_pd3dDevice->SetRenderState( D3DRS_ZENABLE,      FALSE );
				g_pd3dDevice->SetRenderState( D3DRS_ZWRITEENABLE, FALSE );
				g_pd3dDevice->SetRenderState( D3DRS_LIGHTING,     FALSE );

				g_pd3dDevice->SetRenderState( D3DRS_ALPHABLENDENABLE, TRUE );
				g_pd3dDevice->SetRenderState( D3DRS_SRCBLEND,         D3DBLEND_ONE );
				g_pd3dDevice->SetRenderState( D3DRS_DESTBLEND,        D3DBLEND_INVSRCALPHA );

				g_pd3dDevice->SetTexture( 0, g_pHelpQuad1Tex );

				g_pd3dDevice->DrawPrimitiveUP( D3DPT_TRIANGLESTRIP, 2, g_pHelpQuad1, sizeof( TLVERTEX ) );

				g_pd3dDevice->SetTexture( 0, NULL );

				g_pd3dDevice->SetRenderState( D3DRS_ZENABLE,          TRUE  );
				g_pd3dDevice->SetRenderState( D3DRS_ZWRITEENABLE,     TRUE  );
				g_pd3dDevice->SetRenderState( D3DRS_ALPHABLENDENABLE, FALSE );
				g_pd3dDevice->SetRenderState( D3DRS_LIGHTING,         TRUE  );
				break;
			case GAUSSIANCURV:
			case MEANCURV:
				g_pd3dDevice->SetFVF( FVF_TLVERTEX );
				g_pd3dDevice->SetStreamSource( 0, NULL, 0, sizeof( TLVERTEX ) );
				g_pd3dDevice->SetRenderState( D3DRS_ZENABLE,      FALSE );
				g_pd3dDevice->SetRenderState( D3DRS_ZWRITEENABLE, FALSE );
				g_pd3dDevice->SetRenderState( D3DRS_LIGHTING,     FALSE );

				g_pd3dDevice->SetRenderState( D3DRS_ALPHABLENDENABLE, TRUE );
				g_pd3dDevice->SetRenderState( D3DRS_SRCBLEND,         D3DBLEND_ONE );
				g_pd3dDevice->SetRenderState( D3DRS_DESTBLEND,        D3DBLEND_INVSRCALPHA );

				g_pd3dDevice->SetTexture( 0, g_pHelpQuad2Tex );

				g_pd3dDevice->DrawPrimitiveUP( D3DPT_TRIANGLESTRIP, 2, g_pHelpQuad2, sizeof( TLVERTEX ) );

				g_pd3dDevice->SetTexture( 0, NULL );

				g_pd3dDevice->SetRenderState( D3DRS_ZENABLE,          TRUE  );
				g_pd3dDevice->SetRenderState( D3DRS_ZWRITEENABLE,     TRUE  );
				g_pd3dDevice->SetRenderState( D3DRS_ALPHABLENDENABLE, FALSE );
				g_pd3dDevice->SetRenderState( D3DRS_LIGHTING,         TRUE  );
				break;
			case MAGCURVDEV:
				g_pd3dDevice->SetFVF( FVF_TLVERTEX );
				g_pd3dDevice->SetStreamSource( 0, NULL, 0, sizeof( TLVERTEX ) );
				g_pd3dDevice->SetRenderState( D3DRS_ZENABLE,      FALSE );
				g_pd3dDevice->SetRenderState( D3DRS_ZWRITEENABLE, FALSE );
				g_pd3dDevice->SetRenderState( D3DRS_LIGHTING,     FALSE );

				g_pd3dDevice->SetRenderState( D3DRS_ALPHABLENDENABLE, TRUE );
				g_pd3dDevice->SetRenderState( D3DRS_SRCBLEND,         D3DBLEND_ONE );
				g_pd3dDevice->SetRenderState( D3DRS_DESTBLEND,        D3DBLEND_INVSRCALPHA );

				g_pd3dDevice->SetTexture( 0, g_pHelpQuad3Tex );

				g_pd3dDevice->DrawPrimitiveUP( D3DPT_TRIANGLESTRIP, 2, g_pHelpQuad3, sizeof( TLVERTEX ) );

				g_pd3dDevice->SetTexture( 0, NULL );

				g_pd3dDevice->SetRenderState( D3DRS_ZENABLE,          TRUE  );
				g_pd3dDevice->SetRenderState( D3DRS_ZWRITEENABLE,     TRUE  );
				g_pd3dDevice->SetRenderState( D3DRS_ALPHABLENDENABLE, FALSE );
				g_pd3dDevice->SetRenderState( D3DRS_LIGHTING,         TRUE  );
				break;
			}
		}

        // End the scene
        pd3dDevice->EndScene( );
    }
}

////////////////////////////////////////////////////////////////////////////////////////
// Render the help and statistics text. This function uses the ID3DXFont interface for 
// efficient text rendering.
////////////////////////////////////////////////////////////////////////////////////////
void RenderText( )
{
    if( g_iShowUI > 1 )
	{
		g_HUD.SetVisible     ( false );
		g_SampleUI.SetVisible( false );

		return;
	}
	else
	{
		g_HUD.SetVisible     ( true );
		g_SampleUI.SetVisible( true );
	}

    // The helper object simply helps keep track of text position, and color
    // and then it calls pFont->DrawText( m_pSprite, strMsg, -1, &rc, DT_NOCLIP, m_clr );
    // If NULL is passed in as the sprite object, then it will work however the 
    // pFont->DrawText() will not be batched together.  Batching calls will improves 
	// performance.
    const D3DSURFACE_DESC* pd3dsdBackBuffer = DXUTGetBackBufferSurfaceDesc( );
    CDXUTTextHelper txtHelper( g_pFont, g_pTextSprite, 15 );

    // Output statistics
    txtHelper.Begin( );
    txtHelper.SetInsertionPos( 5, 5 );
    txtHelper.SetForegroundColor( D3DXCOLOR( 1.0f, 1.0f, 0.0f, 1.0f ) );
    txtHelper.DrawTextLine( DXUTGetFrameStats( ) );
    txtHelper.DrawTextLine( DXUTGetDeviceStats( ) );

    txtHelper.SetForegroundColor( D3DXCOLOR( 1.0f, 1.0f, 1.0f, 1.0f ) );

    // Output statistics
    switch( g_eVisMode )
    {
        case CURVCOLOR:
            txtHelper.DrawTextLine( L"Showing principal curvatures as colors\n" );
            break;
        case GAUSSIANCURV:
            txtHelper.DrawTextLine( L"Showing Gaussian curvature\n" );
            break;
        case MEANCURV:
            txtHelper.DrawTextLine( L"Showing mean curvature\n" );
            break;
        case PRINCIPALDIRS:
            txtHelper.DrawTextLine( L"Showing principal directions (red=min, green=max)\n" );
            break;
        case MAGCURVDEV:
            txtHelper.DrawTextLine( L"Showing curvature derivative magnitude\n" );
            break;
		case NONE:
        default:
			break;
    }

    // Draw help
    if( g_iShowUI == 0 )
    {
        txtHelper.SetInsertionPos( 10, pd3dsdBackBuffer->Height-15*6 );
        txtHelper.SetForegroundColor( D3DXCOLOR( 1.0f, 0.75f, 0.0f, 1.0f ) );
        txtHelper.DrawTextLine( L"Controls (F1 to hide):" );

        txtHelper.SetInsertionPos( 40, pd3dsdBackBuffer->Height-15*5 );
        txtHelper.DrawTextLine( L"Rotate model: Left click drag\n"
                                L"Zoom: Middle click drag\n"
                                L"Pane: Right click drag\n"
                                L"Quit: ESC" );
    }

    txtHelper.End();
}

////////////////////////////////////////////////////////////////////////////////////////
// Before handling window messages, DXUT passes incoming windows 
// messages to the application through this callback function. If the application sets 
// *pbNoFurtherProcessing to TRUE, then DXUT will not process this message.
////////////////////////////////////////////////////////////////////////////////////////
LRESULT CALLBACK MsgProc( HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam, 
						  bool* pbNoFurtherProcessing, void* pUserContext )
{
    // Always allow dialog resource manager calls to handle global messages
    // so GUI state is updated correctly
    g_DialogResourceManager.MsgProc( hWnd, uMsg, wParam, lParam );

    if( g_SettingsDlg.IsActive() )
    {
        g_SettingsDlg.MsgProc( hWnd, uMsg, wParam, lParam );
        return 0;
    }

    // Give the dialogs a chance to handle the message first
    *pbNoFurtherProcessing = g_HUD.MsgProc( hWnd, uMsg, wParam, lParam );
    if( *pbNoFurtherProcessing )
        return 0;
    *pbNoFurtherProcessing = g_SampleUI.MsgProc( hWnd, uMsg, wParam, lParam );
    if( *pbNoFurtherProcessing )
        return 0;

    // Pass all remaining windows messages to arcball so it can respond to user input
    g_ArcBall.HandleMessages( hWnd, uMsg, wParam, lParam );

    return 0;
}

////////////////////////////////////////////////////////////////////////////////////////
// As a convenience, DXUT inspects the incoming windows messages for
// keystroke messages and decodes the message parameters to pass relevant keyboard
// messages to the application.  The framework does not remove the underlying keystroke 
// messages, which are still passed to the application's MsgProc callback.
////////////////////////////////////////////////////////////////////////////////////////
void CALLBACK KeyboardProc( UINT nChar, bool bKeyDown, bool bAltDown, 
						    void* pUserContext )
{
    if( bKeyDown )
    {
        switch( nChar )
        {
            case VK_F1: 
				g_iShowUI = g_iShowUI + 1; 
				if( g_iShowUI > 3 )
					g_iShowUI = 0;
				break;
        }
    }
}

////////////////////////////////////////////////////////////////////////////////////////
// Handles the GUI events
////////////////////////////////////////////////////////////////////////////////////////
void CALLBACK OnGUIEvent( UINT nEvent, int nControlID, CDXUTControl* pControl, 
						  void* pUserContext )
{
    switch( nControlID )
    {
        case IDC_TOGGLEFULLSCREEN: DXUTToggleFullScreen( ); break;
        case IDC_TOGGLEREF:        DXUTToggleREF( ); break;
        case IDC_CHANGEDEVICE:     g_SettingsDlg.SetActive( !g_SettingsDlg.IsActive( ) ); break;
		case IDC_MESH:
		{
			MESHES NewMesh = (MESHES)(size_t)((CDXUTComboBox*)pControl)->GetSelectedData( );

			if( g_eCurMesh != NewMesh )
			{
				g_eCurMesh = NewMesh;

				DestroyNRTs( );
				DestroyRTs ( );
				ReleaseMesh( );
				LoadMesh   ( );
				CreateRTs  ( );
				CreateNRTs ( );

				D3DSURFACE_DESC pBackBufferSurfaceDesc;
				g_pd3dBackBuffer->GetDesc( &pBackBufferSurfaceDesc );

				// Setup the projection matrix
				g_fZNear = g_fObjectRadius / 64.0f;
				g_fZFar  = g_fObjectRadius * 200.0f;
				D3DXMatrixPerspectiveFovLH( &g_matProj, D3DX_PI / 4, g_fAspect,
					g_fZNear, g_fZFar );
				g_pd3dDevice->SetTransform( D3DTS_PROJECTION, &g_matProj );

				// Setup the arcball parameters
				g_ArcBall.SetWindow( pBackBufferSurfaceDesc.Width, 
					pBackBufferSurfaceDesc.Height, 0.85f );
				g_ArcBall.SetTranslationRadius( g_fObjectRadius );
			}

			break;
		}
        case IDC_METHOD:
        {
            VISMODE NewVisMode = (VISMODE)(size_t)((CDXUTComboBox*)pControl)->GetSelectedData( );

            // If the selected visualization mode is different than the current one
            if( g_eVisMode != NewVisMode )
                g_eVisMode = NewVisMode;
            break;
        }
		case IDC_DIFFUSE:
		{
			g_bUseDiff = ((CDXUTCheckBox*)pControl)->GetChecked( );
			break;
		}
		case IDC_WIREFRAME:
		{
			g_bWireframe = ((CDXUTCheckBox*)pControl)->GetChecked( );
			break;
		}
    }
}

////////////////////////////////////////////////////////////////////////////////////////
// This callback function will be called immediately after the Direct3D device has 
// entered a lost state and before IDirect3DDevice9::Reset is called. Resources created
// in the OnResetDevice callback should be released here, which generally includes all 
// D3DPOOL_DEFAULT resources. See the "Lost Devices" section of the documentation for 
// information about lost devices.
////////////////////////////////////////////////////////////////////////////////////////
void CALLBACK OnLostDevice( void* pUserContext )
{
    g_DialogResourceManager.OnLostDevice( );
    g_SettingsDlg.OnLostDevice( );
    if( g_pFont )
        g_pFont->OnLostDevice( );
    if( g_pEffect )
        g_pEffect->OnLostDevice( );
    SAFE_RELEASE( g_pTextSprite );

	DestroyShaders( );
	DestroyNRTs   ( );
	DestroyRTs    ( );
}

////////////////////////////////////////////////////////////////////////////////////////
// This callback function will be called immediately after the Direct3D device has 
// been destroyed, which generally happens as a result of application termination or 
// windowed/full screen toggles. Resources created in the OnCreateDevice callback 
// should be released here, which generally includes all D3DPOOL_MANAGED resources. 
////////////////////////////////////////////////////////////////////////////////////////
void CALLBACK OnDestroyDevice( void* pUserContext )
{
	ReleaseMesh( );

	SAFE_RELEASE( g_pHelpQuad1Tex );
	SAFE_RELEASE( g_pHelpQuad2Tex );
	SAFE_RELEASE( g_pHelpQuad3Tex );

    g_DialogResourceManager.OnDestroyDevice( );
    g_SettingsDlg.OnDestroyDevice( );
	SAFE_RELEASE( g_pMesh   );
    SAFE_RELEASE( g_pEffect );
    SAFE_RELEASE( g_pFont   );
}

////////////////////////////////////////////////////////////////////////////////
// Create the render target textures for storing the geometry attributes.
////////////////////////////////////////////////////////////////////////////////
HRESULT CreateRTs( )
{
	// Set up data for quad rendering
    g_pScreenQuadVt[0].p = D3DXVECTOR4(                 - 0.5f,                 - 0.5f, 0.0f, 1.0f );
    g_pScreenQuadVt[1].p = D3DXVECTOR4( g_iVtAttTexSize - 0.5f,                 - 0.5f, 0.0f, 1.0f );
	g_pScreenQuadVt[2].p = D3DXVECTOR4(                 - 0.5f, g_iVtAttTexSize - 0.5f, 0.0f, 1.0f );
    g_pScreenQuadVt[3].p = D3DXVECTOR4( g_iVtAttTexSize - 0.5f, g_iVtAttTexSize - 0.5f, 0.0f, 1.0f );

    g_pScreenQuadVt[0].t = D3DXVECTOR2( 0.0f, 0.0f );
    g_pScreenQuadVt[1].t = D3DXVECTOR2( 1.0f, 0.0f );
    g_pScreenQuadVt[2].t = D3DXVECTOR2( 0.0f, 1.0f );
    g_pScreenQuadVt[3].t = D3DXVECTOR2( 1.0f, 1.0f );

	// Create 'vertex position' texture

	// Create render target texture
	HRESULT hr;
	V_RETURN( g_pd3dDevice->CreateTexture( g_iVtAttTexSize, g_iVtAttTexSize,
		1, D3DUSAGE_RENDERTARGET, D3DFMT_A32B32G32R32F,D3DPOOL_DEFAULT, 
		&g_pVtPosTex, NULL ) );

	// In order to fill the vertex position texture, we need to
	// create a temporary texture in system memory, which is uploaded
	// to video memory by using IDirect3DDevice9::UpdateTexture

	// Create temporary system memory texture
	LPDIRECT3DTEXTURE9 pd3dTempTex;
	V_RETURN( g_pd3dDevice->CreateTexture( g_iVtAttTexSize, g_iVtAttTexSize,
		1, 0, D3DFMT_A32B32G32R32F, D3DPOOL_SYSTEMMEM, &pd3dTempTex, NULL ) );

	D3DLOCKED_RECT rcLockedRect;
	if( SUCCEEDED( pd3dTempTex->LockRect( 0, &rcLockedRect, NULL, 0 ) ) )
	{
		int iVert = 0;
		for( int iY = 0; iY < g_iVtAttTexSize; iY++ )
		{
			for( int iX = 0; iX < g_iVtAttTexSize; iX++ )
			{
				BYTE *pLockedRectB = (BYTE*) rcLockedRect.pBits;
				float *vTexel = (float*) ( pLockedRectB + 
					(iX*4*sizeof(float) + iY*rcLockedRect.Pitch ) );

				if( iVert >= g_iNumVt )
					// Null texels
					vTexel[0] = vTexel[1] = vTexel[2] = vTexel[3] = -1.0f;
				else
				{
					// Get vertex position for this texel
					vTexel[0] = g_pVtPos[iVert*3+0];
					vTexel[1] = g_pVtPos[iVert*3+1];
					vTexel[2] = g_pVtPos[iVert*3+2];
				}

				// Go to next texel
				iVert++;
			}
		}
		pd3dTempTex->UnlockRect( 0 );
	}

	// Copy from system memory to video memory
	V_RETURN( g_pd3dDevice->UpdateTexture( pd3dTempTex, g_pVtPosTex ) );

	// Create 'vertex normal' render target texture
	V_RETURN( g_pd3dDevice->CreateTexture( g_iVtAttTexSize, g_iVtAttTexSize, 
		1, D3DUSAGE_RENDERTARGET, D3DFMT_A32B32G32R32F,	D3DPOOL_DEFAULT, 
		&g_pVtNormalTex, NULL ) );

	// In order to fill the vertex normal texture, we need to
	// create a temporary texture in system memory, which is uploaded
	// to video memory by using IDirect3DDevice9::UpdateTexture

	// Fill texture
	if( SUCCEEDED( pd3dTempTex->LockRect( 0, &rcLockedRect, NULL, 0 ) ) )
	{
		int iVert = 0;
		for( int iY = 0; iY < g_iVtAttTexSize; iY++ )
		{
			for( int iX = 0; iX < g_iVtAttTexSize; iX++ )
			{
				BYTE *pLockedRectB = (BYTE*) rcLockedRect.pBits;
				float *vTexel = (float*) ( pLockedRectB + 
					(iX*4*sizeof(float) + iY*rcLockedRect.Pitch ) );

				if( iVert >= g_iNumVt )
					// Null texels
					vTexel[0] = vTexel[1] = vTexel[2] = vTexel[3] = -1.0f;
				else
				{
					// Get vertex normal for this texel
					vTexel[0] = g_pVtNormal[iVert*3+0];
					vTexel[1] = g_pVtNormal[iVert*3+1];
					vTexel[2] = g_pVtNormal[iVert*3+2];
				}

				// Go to next texel
				iVert++;
			}
		}
		pd3dTempTex->UnlockRect( 0 );
	}

	// Copy from system memory to video memory
	V_RETURN( g_pd3dDevice->UpdateTexture( pd3dTempTex, g_pVtNormalTex ) );

	SAFE_RELEASE( pd3dTempTex );

	// Create render target texture for tensor of curvature
	V_RETURN( g_pd3dDevice->CreateTexture( g_iVtAttTexSize, g_iVtAttTexSize, 
		1, D3DUSAGE_RENDERTARGET, D3DFMT_A32B32G32R32F, D3DPOOL_DEFAULT, 
		&g_pVtTensor2Tex, NULL ) );

	// Create render target texture for #1 principal direction and curvature
	V_RETURN( g_pd3dDevice->CreateTexture( g_iVtAttTexSize, g_iVtAttTexSize, 
		1, D3DUSAGE_RENDERTARGET, D3DFMT_A32B32G32R32F, D3DPOOL_DEFAULT, 
		&g_pVtP1K1Tex, NULL ) );

	// Create render target texture for #2 principal direction and curvature
	V_RETURN( g_pd3dDevice->CreateTexture( g_iVtAttTexSize, g_iVtAttTexSize, 
		1, D3DUSAGE_RENDERTARGET, D3DFMT_A32B32G32R32F, D3DPOOL_DEFAULT, 
		&g_pVtP2K2Tex, NULL ) );

	// Create render target texture for tensor of derivative of curvature
	V_RETURN( g_pd3dDevice->CreateTexture( g_iVtAttTexSize, g_iVtAttTexSize, 
		1, D3DUSAGE_RENDERTARGET, D3DFMT_A32B32G32R32F,	D3DPOOL_DEFAULT, 
		&g_pVtTensor3Tex, NULL ) );

	// Get back buffer because we are supposed to switch to it after rendering
	// to the off-screen render targets
	V_RETURN( g_pd3dDevice->GetRenderTarget( 0, &g_pd3dBackBuffer ) );

	// Create temporary surface to read vertex attributes
	V_RETURN( g_pd3dDevice->CreateOffscreenPlainSurface( g_iVtAttTexSize, 
		g_iVtAttTexSize, D3DFMT_A32B32G32R32F, D3DPOOL_SYSTEMMEM, &g_pReadSurf, 
		NULL ) );

	return S_OK;
}

////////////////////////////////////////////////////////////////////////////////
// Release the render target textures created with CreateRTs.
////////////////////////////////////////////////////////////////////////////////
HRESULT DestroyRTs( )
{
	// Release temporary read surface
	SAFE_RELEASE( g_pReadSurf );

	// Release the reference count to the back buffer
	SAFE_RELEASE( g_pd3dBackBuffer );

	// Release render textures with vertex attributes
	SAFE_RELEASE( g_pVtPosTex     );
	SAFE_RELEASE( g_pVtNormalTex  );
	SAFE_RELEASE( g_pVtTensor2Tex );
	SAFE_RELEASE( g_pVtP1K1Tex    );
	SAFE_RELEASE( g_pVtP2K2Tex    );
	SAFE_RELEASE( g_pVtTensor3Tex );

	return S_OK;
}

////////////////////////////////////////////////////////////////////////////////
// Create the non-render-target textures used as data structures for 
// computing attributes of interaction.
////////////////////////////////////////////////////////////////////////////////
HRESULT CreateNRTs( )
{
	int         *pAdjFaces    = NULL;
	vector<int> *pAdjVtsSTL   = NULL;
	int          iNumAdjVt    = 0;
	int         *pAdjVts      = NULL;
	int			*pAdjAddress  = NULL;
	
	g_iAdjVtTexSize = 0;

	// Create auxiliary data structures used for creating the textures with adjacency data
	if( g_pFaceIdx == NULL )
		return E_FAIL;

	// Create 'adjacent vertices' temporary STL set
	pAdjVtsSTL = new vector<int>[ g_iNumVt ];
	if( pAdjVtsSTL == NULL )
		return E_OUTOFMEMORY;

	for( int iFace = 0; iFace < g_iNumFace; iFace++ )
	{
		int iIdx, iAdd;
		int iFaceIdx = iFace*3;

		for( int iVt = 0; iVt < 3; iVt++ )
		{
			iIdx = g_pFaceIdx[iFaceIdx+iVt];
			vector<int> &vCur = pAdjVtsSTL[iIdx];
			iAdd = g_pFaceIdx[iFaceIdx+((iVt+1)%3)];
			if( find( vCur.begin( ), vCur.end( ), iAdd ) == vCur.end( ) )
				vCur.push_back( iAdd );
			iAdd = g_pFaceIdx[iFaceIdx+((iVt+2)%3)];
			if( find( vCur.begin( ), vCur.end( ), iAdd ) == vCur.end( ) )
				vCur.push_back( iAdd );
		}
	}

	// Determine size of 'adjacent vertices' texture
	for( int iVt = 0; iVt < g_iNumVt; iVt++ )
		iNumAdjVt += (int) pAdjVtsSTL[iVt].size( );

	// Compute minimum size required by the 'adjacent vertices' texture
	DWORD dwPow = 1;
	g_iAdjVtTexSize = 0;
	while( 1 )
	{
		g_iAdjVtTexSize = 1 << dwPow++;
		if( g_iAdjVtTexSize * g_iAdjVtTexSize >= iNumAdjVt )
			break;
	}

	// Now create an array of adjacent vertices in a format which 
	// is easier to assign to a texture further
	pAdjVts = new int[ iNumAdjVt ];
	if( pAdjVts == NULL )
		return E_OUTOFMEMORY;
	int iCt = 0;
	for( int iVt = 0; iVt < g_iNumVt; iVt++ )
	{
		vector<int>::iterator stlIt = pAdjVtsSTL[iVt].begin( );
		while( stlIt != pAdjVtsSTL[iVt].end( ) )
		{
			pAdjVts[iCt] = (*stlIt);

			iCt++;
			stlIt++;
		}
	}

	// Create temporary array of addresses of the beginning of the
	// array of adjacent vertices.
	// Each group of 2 elements corresponds to a vertex and contains the 
	// following values:
	// 1) Address to the 1st adjacent vertex in the vertex adjacency 
	//    texture
	// 2) Number of adjacent vertices
	pAdjAddress = new int[ g_iNumVt * 2 ];
	if( pAdjAddress == NULL )
		return -1;
	memset( pAdjAddress, 0, sizeof( int ) * g_iNumVt * 2 );

	int iCtVt = 0;
	for( int iVt = 0; iVt < g_iNumVt; iVt++ )
	{
		pAdjAddress[iVt*2+0] = iCtVt;
		pAdjAddress[iVt*2+1] = (int) pAdjVtsSTL[iVt].size( );

		iCtVt   += (int) pAdjVtsSTL[iVt].size( );
	}

	// Create 'adjacent vertices' texture
	HRESULT hr;
	V_RETURN( g_pd3dDevice->CreateTexture( g_iAdjVtTexSize,	g_iAdjVtTexSize,
		1, 0, D3DFMT_R32F, D3DPOOL_MANAGED, &g_pAdjVtTex, NULL ) );

	D3DLOCKED_RECT rcLockedRect;
	if( SUCCEEDED( g_pAdjVtTex->LockRect( 0, &rcLockedRect, NULL, 0 ) ) )
	{
		int iVtIdx = 0;
		for( int iY = 0; iY < g_iAdjVtTexSize; iY++ )
		{
			for( int iX = 0; iX < g_iAdjVtTexSize; iX++ )
			{
				BYTE *pLockedRectB = (BYTE*) rcLockedRect.pBits;
				float *vTexel = (float*) ( pLockedRectB + 
					(iX*1*sizeof(float) + iY*rcLockedRect.Pitch ) );

				vTexel[0] = ( iVtIdx >= iNumAdjVt ) ? -1.0f : (float) pAdjVts[iVtIdx]; 

				iVtIdx++;
			}
		}
		g_pAdjVtTex->UnlockRect( 0 );
	}

	// Create 'adjacency addresses' texture
	V_RETURN( g_pd3dDevice->CreateTexture( g_iVtAttTexSize, g_iVtAttTexSize, 
		1, 0, D3DFMT_G32R32F, D3DPOOL_MANAGED, &g_pAdjAddTex, NULL ) );

	if( SUCCEEDED( g_pAdjAddTex->LockRect( 0, &rcLockedRect, NULL, 0 ) ) )
	{
		int iVtIdx = 0;
		for( int iY = 0; iY < g_iVtAttTexSize; iY++ )
		{
			for( int iX = 0; iX < g_iVtAttTexSize; iX++ )
			{
				BYTE *pLockedRectB = (BYTE*) rcLockedRect.pBits;
				float *vTexel = (float*) ( pLockedRectB + 
					(iX*2*sizeof(float) + iY*rcLockedRect.Pitch ) );

				if( iVtIdx >= g_iNumVt )
				{
					// Null texels
					vTexel[0] = vTexel[1] = -1.0f;
					iX = iY = g_iVtAttTexSize;
				}
				else
				{
					vTexel[0] = (float) pAdjAddress[iVtIdx*2+0];
					vTexel[1] = (float) pAdjAddress[iVtIdx*2+1];
				}

				// Go to next texel
				iVtIdx++;
			}
		}
		g_pAdjAddTex->UnlockRect( 0 );
	}

	SAFE_DELETE_ARRAY( pAdjFaces    );
	SAFE_DELETE_ARRAY( pAdjVtsSTL   );
	SAFE_DELETE_ARRAY( pAdjVts      );
	SAFE_DELETE_ARRAY( pAdjAddress  );

	return S_OK;
}

////////////////////////////////////////////////////////////////////////////////
// Release the textures created with CreateNRTs.
////////////////////////////////////////////////////////////////////////////////
HRESULT DestroyNRTs( )
{
	// Release textures with vertex adjacency data
	SAFE_RELEASE( g_pAdjVtTex   );
	SAFE_RELEASE( g_pAdjAddTex  );

	return S_OK;
} // DestroyNRTs

////////////////////////////////////////////////////////////////////////////////
// Create the shaders.
////////////////////////////////////////////////////////////////////////////////
HRESULT CreateShaders( )
{
	HRESULT hr;

	// Define DEBUG_VS and/or DEBUG_PS to debug vertex and/or pixel shaders 
	// with the shader debugger. Debugging vertex shaders requires either REF 
	// or software vertex processing, and debugging pixel shaders requires REF.  
	// The D3DXSHADER_FORCE_*_SOFTWARE_NOOPT flag improves the debug experience 
	// in the shader debugger.  It enables source level debugging, prevents 
	// instruction reordering, prevents dead code elimination, and forces the 
	// compiler to compile against the next higher available software target, 
	// which ensures that the unoptimized shaders do not exceed the shader model 
	// limitations.  Setting these flags will cause slower rendering since the 
	// shaders will be unoptimized and forced into software.  See the DirectX 
	// documentation for more information about using the shader debugger.
	DWORD dwShaderFlags = 0;
#ifdef DEBUG_VS
	dwShaderFlags |= D3DXSHADER_FORCE_VS_SOFTWARE_NOOPT;
#endif
#ifdef DEBUG_PS
	dwShaderFlags |= D3DXSHADER_FORCE_PS_SOFTWARE_NOOPT;
#endif

	WCHAR str[MAX_PATH];

	// Load effect for computing diff. geometry properties
	V_RETURN( DXUTFindDXSDKMediaFileCch( str, MAX_PATH, L"GeomDiff.fx" ) );
	LPD3DXBUFFER pd3dBuffer = NULL;
	V_RETURN( D3DXCreateEffectFromFile( g_pd3dDevice, str,
		NULL, NULL, dwShaderFlags, NULL, &g_pGDFx, &pd3dBuffer ) );

	// Get handle to uniform parameters in the effect
	g_hGDAdjVtIdxC = g_pGDFx->GetParameterByName( NULL, "g_fAdjVtIdxC" );
	g_hGDVtIdxC    = g_pGDFx->GetParameterByName( NULL, "g_fVtIdxC"    );

	// Load effect for rendering the geometry
	V_RETURN( DXUTFindDXSDKMediaFileCch( str, MAX_PATH, L"Render.fx" ) );
	V_RETURN( D3DXCreateEffectFromFile( g_pd3dDevice, str,
		NULL, NULL, dwShaderFlags, NULL, &g_pRenderFx, &pd3dBuffer ) );

	// Get handle to uniform parameters in the render effect
	g_hRenderFxIdxC = g_pRenderFx->GetParameterByName( NULL, "g_fVtIdxC"  );
	g_hRenderFxIW   = g_pRenderFx->GetParameterByName( NULL, "g_mIW"      );
	g_hRenderFxWVP  = g_pRenderFx->GetParameterByName( NULL, "g_mWVP"     );
	g_hRenderFxITW  = g_pRenderFx->GetParameterByName( NULL, "g_mITW"     );
	g_hRenderFxDiff = g_pRenderFx->GetParameterByName( NULL, "g_bUseDiff" );

	return S_OK;
} // CreateShaders

////////////////////////////////////////////////////////////////////////////////
// Release the shaders created with CreateShaders.
////////////////////////////////////////////////////////////////////////////////
HRESULT DestroyShaders( )
{
	SAFE_RELEASE( g_pRenderFx );
	SAFE_RELEASE( g_pGDFx     );

	return S_OK;
} // DestroyShaders