////////////////////////////////////////////////////////////////////////////
// File: GeomDiff.fx
//
// Effect for the estimation of discrete differential geometry properties
// from triangle meshes.
//
// Copyright (C) 2007 Harlen Costa Batagelo.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// Lesser General Public License for more details.
////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////
// Global variables
////////////////////////////////////////////////////////////////////////////
float2 g_fAdjVtIdxC; // 1D-2D constant (vertex adjacency)
float2 g_fVtIdxC;    // 1D-2D constant (vertex attributes)

////////////////////////////////////////////////////////////////////////////
// Vertex shader input semantics
////////////////////////////////////////////////////////////////////////////
struct VS_IN
{
	float4 vPos : POSITION;  // Transformed position (XYZ+RHW)
    float2 vTex : TEXCOORD0; // Texture coordinates (UV).	
};

// Vertex position
texture Tex0;
sampler s0 = sampler_state
{
    Texture   = (Tex0);
	MipFilter = NONE; MinFilter = POINT; MagFilter = POINT;
    AddressU  = Wrap; AddressV  = Wrap;
};

// Per-vertex normal
texture Tex1;
sampler s1 = sampler_state
{
    Texture   = (Tex1);
	MipFilter = NONE; MinFilter = POINT; MagFilter = POINT;
    AddressU  = Wrap; AddressV  = Wrap;
};

// Adjacency address
texture Tex2;
sampler s2 = sampler_state
{
    Texture   = (Tex2);
	MipFilter = NONE; MinFilter = POINT; MagFilter = POINT;
    AddressU  = Wrap; AddressV  = Wrap;
};

// Vertex adjacency
texture Tex3;
sampler s3 = sampler_state
{
    Texture   = (Tex3);
	MipFilter = NONE; MinFilter = POINT; MagFilter = POINT;
    AddressU  = Wrap; AddressV  = Wrap;
};

// #1 principal direction and curvature
texture Tex4;
sampler s4 = sampler_state
{
    Texture   = (Tex4);
	MipFilter = NONE; MinFilter = POINT; MagFilter = POINT;
    AddressU  = Wrap; AddressV  = Wrap;
};

// #2 principal direction and curvature
texture Tex5;
sampler s5 = sampler_state
{
    Texture   = (Tex5);
	MipFilter = NONE; MinFilter = POINT; MagFilter = POINT;
    AddressU  = Wrap; AddressV  = Wrap;
};

////////////////////////////////////////////////////////////////////////////
// Pixel shader output semantics
////////////////////////////////////////////////////////////////////////////
struct PS_OUT
{
	// vColor[0]: Tensor of curvature (xyz)
	// vColor[1]: #1 principal direction (xyz) and curvature (w)
	// vColor[2]: #2 principal direction (xyz) and curvature (w)
	float4 vColor[3] : COLOR;
};

////////////////////////////////////////////////////////////////////////////
// Perform LDL^T decomposition of a 3x3 symmetric positive definite matrix
// and solves Ax=B.
////////////////////////////////////////////////////////////////////////////
float3 ldltsv3( float3x3 w, float3 m )
{
	// Decompose matrix. Overwrites lower triangle of matrix.
	float3 d;
	float2 v;
	int i;
	for( i = 0; i < 3; i++ ) 
	{
		for( int k = 0; k < i; k++ ) v[k] = w[i][k] * d[k];
		for( int j = i; j < 3; j++ ) 
		{
			float sum = w[i][j];
			for( int k = 0; k < i; k++ ) 
				sum -= v[k] * w[j][k];
			if( i == j ) 
				d[i] = 1.0f / sum;
			else
				w[j][i] = sum;
		}
	}
	
	// Solve Ax=B after LDL^T decomposition
	for( i = 0; i < 3; i++ )
	{
		float sum = m[i];
		for( int k = 0; k < i; k++ ) 
			sum -= w[i][k] * m[k];
		m[i] = sum * d[i];
	}
	for( i = 2; i >= 0; i-- )
	{
		float sum = 0;
		for( int k = i + 1; k < 3; k++ ) 
			sum += w[k][i] * m[k];
		m[i] -= sum * d[i];
	}
	
	return m;
}

////////////////////////////////////////////////////////////////////////////
// Perform LDL^T decomposition of a 4x4 symmetric positive definite matrix
// and solves Ax=B.
////////////////////////////////////////////////////////////////////////////
float4 ldltsv4( float4x4 w, float4 m )
{
	// Decompose matrix. Overwrites lower triangle of matrix.
	float4 d;
	float3 v;
	int i;
	for( i = 0; i < 4; i++ ) 
	{
		for( int k = 0; k < i; k++ ) v[k] = w[i][k] * d[k];
		for( int j = i; j < 4; j++ ) 
		{
			float sum = w[i][j];
			for( int k = 0; k < i; k++ ) 
				sum -= v[k] * w[j][k];
			if( i == j ) 
				d[i] = 1.0f / sum;
			else
				w[j][i] = sum;
		}
	}
	
	// Solve Ax=B after LDL^T decomposition
	for( i = 0; i < 4; i++ )
	{
		float sum = m[i];
		for( int k = 0; k < i; k++ ) 
			sum -= w[i][k] * m[k];
		m[i] = sum * d[i];
	}
	for( i = 3; i >= 0; i-- )
	{
		float sum = 0;
		for( int k = i + 1; k < 4; k++ ) 
			sum += w[k][i] * m[k];
		m[i] -= sum * d[i];
	}
	
	return m;
}

////////////////////////////////////////////////////////////////////////////
/// \brief Convert 1D index (i) to 2D UV index.
/// c = {1/W, 1/(W*H)} where W and H are the width and height of the
///     texture map
////////////////////////////////////////////////////////////////////////////
float2 I1DTo2D( float i, float2 c ) { return float2( i, i ) * c; } 

////////////////////////////////////////////////////////////////////////////
// Pixel shader for computing the tensor of curvature, the principal
// directions and principal curvatures.
////////////////////////////////////////////////////////////////////////////
PS_OUT Tensor2PS( VS_IN In )
{
  PS_OUT Out;
	
  // Sample adjacency addresses
  // .x) Address to the 1st adjacent vertex in the vertex adjacency 
  //     texture
  // .y) Number of adjacent vertices
  float2 fAdj = tex2D( s2, In.vTex );
  if( fAdj.x == -1.0 ) discard;
	
  // UV index to the first adjacent vertex
  float2 vAdjUVNei  = I1DTo2D( fAdj.x, g_fAdjVtIdxC );
  float  fIdxNei    = tex2D( s3, vAdjUVNei ).x;	
  float2 vUVNei     = I1DTo2D( fIdxNei, g_fVtIdxC );
	
  float3 vPos       = tex2D( s0, In.vTex );
  float3 vNormal    = tex2D( s1, In.vTex ); 
  float3 vPosNei    = tex2D( s0, vUVNei  );
  float3 vNormalNei = tex2D( s1, vUVNei  );
  float3 vPDir1     = normalize( cross( (vPosNei - vPos), vNormal ) );
  float3 vPDir2     = cross( vNormal, vPDir1 );
	
  float3 m = 0;	
  float3x3 w = 0;
	
  int iCt = 0;
  while( iCt < fAdj.y )
  {
	// Get neighbor index from the vertex adjacency texture
	vAdjUVNei  = I1DTo2D( fAdj.x + iCt, g_fAdjVtIdxC );		
	fIdxNei    = tex2Dlod( s3, float4( vAdjUVNei,0,0 ) ).x;
		
	// Get attributes of neighbor vertex		
	vUVNei     = I1DTo2D( fIdxNei, g_fVtIdxC );		
	vPosNei    = tex2Dlod( s0, float4( vUVNei,0,0 ) );
	vNormalNei = tex2Dlod( s1, float4( vUVNei,0,0 ) );		
    
	float3 e = vPosNei - vPos;
	float  u = dot( e, vPDir1 );
	float  v = dot( e, vPDir2 );
	w[0][0] += u*u;
	w[0][1] += u*v;
	w[2][2] += v*v;
	float3 dn = vNormalNei - vNormal;
	float dnu = dot( dn, vPDir1 );
	float dnv = dot( dn, vPDir2 );
	m[0] += dnu*u;
	m[1] += dnu*v + dnv*u;
	m[2] += dnv*v;
	
	iCt++;		
  } 
    
  w[1][1] = w[0][0] + w[2][2];
  w[1][2] = w[0][1];
    
  m = ldltsv3( w, m );
  Out.vColor[0] = float4( m, 0 );	

  // Jacobi rotation to diagonalize
  float ku = m[0], kuv = m[1], kv = m[2];
  float c = 1, s = 0, tt = 0;
  if( kuv != 0.0f ) 
  {
	float h = 0.5f * (kv - ku) / kuv;
	tt = (h < 0.0f) ? 1.0f / (h - (float) sqrt(1.0f + h*h)) :
		 1.0f / (h + (float) sqrt(1.0f + h*h));
	c = 1.0f / (float) sqrt(1.0f + tt*tt);
	s = tt * c;
  }

  float fK1 = ku - tt * kuv, 
	    fK2 = kv + tt * kuv;

  float3 pd1, pd2;
  if( abs( fK1 ) >= abs( fK2 ) )
	pd1 = c * vPDir1 - s * vPDir2;
  else 
  {
	float fTemp = fK1; fK1 = fK2; fK1 = fTemp;
 	pd1 = s * vPDir1 + c * vPDir2;
  }

  // #1 principal direction and curvature
  Out.vColor[1] = float4( pd1, fK1 );

  // #2 principal direction and curvature
  Out.vColor[2] = float4( cross( vNormal, pd1 ), fK2 );

  return Out;
}

////////////////////////////////////////////////////////////////////////////
// Pixel shader for computing the tensor of curvature derivative.
////////////////////////////////////////////////////////////////////////////
float4 Tensor3PS( VS_IN In ) : COLOR
{
  // Sample adjacency addresses
  // .x) Address to the 1st adjacent vertex in the vertex adjacency 
  //     texture
  // .y) Number of adjacent vertices
  float2 fAdj = tex2D( s2, In.vTex );
  if( fAdj.x == -1.0 ) discard;
	
  float3 vPos   = tex2D( s0, In.vTex );	
  float4 vpk1   = tex2D( s4, In.vTex ); 
  float4 vpk2   = tex2D( s5, In.vTex ); 	
  float3 vPDir1 = vpk1.xyz, vPDir2 = vpk2.xyz;
  float2 vCurv  = float2( vpk1.w, vpk2.w );
  float2 vCurvNei;

  float4 m = 0;	
  float4x4 w = 0;
	
  int iCt = 0;
  while( iCt < fAdj.y )
  {
	// Get neighbor index from the vertex adjacency texture
	float2 vAdjUVNei = I1DTo2D( fAdj.x + iCt, g_fAdjVtIdxC );		
	float  fIdxNei   = tex2Dlod( s3, float4( vAdjUVNei,0,0 ) ).x;
		
	// Get attributes of neighbor vertex		
	float2 vUVNei    = I1DTo2D( fIdxNei, g_fVtIdxC );		
	float3 vPosNei   = tex2Dlod( s0, float4( vUVNei,0,0 ) );
	vCurvNei.x       = tex2Dlod( s4, float4( vUVNei,0,0 ) ).w;
	vCurvNei.y       = tex2Dlod( s5, float4( vUVNei,0,0 ) ).w;
    
	float3 fcurv_i = float3( vCurv.x,    0, vCurv.y    ), 
	       fcurv_n = float3( vCurvNei.x, 0, vCurvNei.y );
    
	// Variation of curvature along each edge
	float3 dfcurv = fcurv_n - fcurv_i;
	float3 e = vPosNei - vPos;

	float u = dot( e, vPDir1 );
	float v = dot( e, vPDir2 );
	float u2 = u*u, v2 = v*v, uv = u*v;
	w[0][0] += u2;
	w[0][1] += uv;
	w[3][3] += v2;
	m[0] += u*dfcurv[0];
	m[3] += v*dfcurv[2];
    	
	iCt++;
  } 
    
  w[1][1] = 2.0f * w[0][0] + w[3][3];
  w[1][2] = 2.0f * w[0][1];
  w[2][2] = w[0][0] + 2.0f * w[3][3];
  w[2][3] = w[0][1];
    
  // Four non-symmetric components of the tensor of curvature derivative
  return ldltsv4( w, m );
}

////////////////////////////////////////////////////////////////////////////
// Effect technique for computing per-vertex curvature tensors
////////////////////////////////////////////////////////////////////////////
technique Tensor2
{
  pass P0
  {          
    VertexShader = null;
    PixelShader  = compile ps_3_0 Tensor2PS( );
  }
}

////////////////////////////////////////////////////////////////////////////
// Effect technique for computing per-vertex deriv. of curvature tensors
////////////////////////////////////////////////////////////////////////////
technique Tensor3
{
  pass P0
  {          
    VertexShader = null;
    PixelShader  = compile ps_3_0 Tensor3PS( );
  }
}