////////////////////////////////////////////////////////////////////////////
// File: Render.fx
//
// Shaders used to visualize the estimated differential geometry properties.
//
// Copyright (C) 2007 Harlen Costa Batagelo.
//
// This library is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 2.1 of the License, or (at your option) any later version.
//
// This library is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
// Lesser General Public License for more details.
////////////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////////////
// Global variables
////////////////////////////////////////////////////////////////////////////
float2   g_fVtIdxC;  // 1D-2D constant (vertex attributes)
float4x4 g_mIW;      // World
float4x4 g_mWVP;     // World*view*proj
float4x4 g_mITW;     // Inverse transpose of the world
bool     g_bUseDiff; // Whether to use diffuse lighting

////////////////////////////////////////////////////////////////////////////
// Constants
////////////////////////////////////////////////////////////////////////////
#define M_2_PI 0.63661977236758134307 // 2.0f / PI
#define M_PI_2 1.57079632679489661923 // PI / 2.0f
#define M_PI   3.14159265358979323846 // PI

// Vertex position
texture Tex0;
sampler s0 = sampler_state
{
  Texture   = (Tex0);
  MipFilter = NONE; MinFilter = POINT; MagFilter = POINT;
  AddressU  = Wrap; AddressV  = Wrap;
};

// Vertex normal
texture Tex1;
sampler s1 = sampler_state
{
  Texture   = (Tex1);
  MipFilter = NONE; MinFilter = POINT; MagFilter = POINT;
  AddressU  = Wrap; AddressV  = Wrap;
};

// #1 principal direction and curvature
texture Tex2;
sampler s2 = sampler_state
{
  Texture   = (Tex2);
  MipFilter = NONE; MinFilter = POINT; MagFilter = POINT;
  AddressU  = Wrap; AddressV  = Wrap;
};

// #2 principal direction and curvature
texture Tex3;
sampler s3 = sampler_state
{
  Texture   = (Tex3);
  MipFilter = NONE; MinFilter = POINT; MagFilter = POINT;
  AddressU  = Wrap; AddressV  = Wrap;
};

// Tensor of derivative of curvature
texture Tex4;
sampler s4 = sampler_state
{
  Texture   = (Tex4);
  MipFilter = NONE; MinFilter = POINT; MagFilter = POINT;
  AddressU  = Wrap; AddressV  = Wrap;
};

////////////////////////////////////////////////////////////////////////////
// Vertex shader input/output semantics.
////////////////////////////////////////////////////////////////////////////
struct VS_IN
{
	float4 vP  : POSITION;  // Position in object space
	float3 vN  : NORMAL;    // Normal vector
    float  fID : TEXCOORD0; // Vertex ID
};

struct VS_OUT
{
	float4 vP     : POSITION; // Transformed position (XYZ+RHW)
    float4 vColor : COLOR;    // Vertex color
};

////////////////////////////////////////////////////////////////////////////
// Convert 1D index (i) to 2D UV index.
// c = {1/W, 1/(W*H)} where W and H are the width and height of the
//     texture map
////////////////////////////////////////////////////////////////////////////
float2 I1DTo2D( float i, float2 c ) { return float2( i, i ) * c; } 

////////////////////////////////////////////////////////////////////////////
// Returns the RGB color from an input color in HSV space
////////////////////////////////////////////////////////////////////////////
float3 hsv2rgb( float3 hsv )
{
	// From FvD
	hsv.x = fmod(hsv.x, 2.0 * M_PI);
	if (hsv.x < 0.0) hsv.x += 2.0 * M_PI;
	hsv.x /= M_PI / 3.0;
	int i = int(floor(hsv.x));
	float f = hsv.x - i;
	float3 pqt = float3(hsv.z * (1.0 - hsv.y), 
						hsv.z * (1.0 - (hsv.y*f)), 
						hsv.z * (1.0 - (hsv.y*(1.0-f))) );
	if     ( i == 0 ) return float3(hsv.z, pqt.z, pqt.x);
	else if( i == 1 ) return float3(pqt.y, hsv.z, pqt.x);
	else if( i == 2 ) return float3(pqt.x, hsv.z, pqt.z);
	else if( i == 3 ) return float3(pqt.x, pqt.y, hsv.z);
	else if( i == 4 ) return float3(pqt.z, pqt.x, hsv.z);
	else              return float3(hsv.z, pqt.x, pqt.y);
}

#define SCALE 64.0

////////////////////////////////////////////////////////////////////////////
// Vertex shader.
////////////////////////////////////////////////////////////////////////////
VS_OUT RenderVS( VS_IN In, uniform int iVisMode )
{
  VS_OUT Out;

  // UV coordinates of the vertex attribute texture
  float2 vVtUV       = I1DTo2D( In.fID, g_fVtIdxC );
  float3 vSampledPos = tex2Dlod( s0, float4( vVtUV, 0, 0 ) );

  // Transform deformed position
  Out.vP = mul( float4( vSampledPos, 1.0f ), g_mWVP );  

  // Compute diffuse color  
  float3 vDiff = 1.0f;
  float4 vGD   = 1.0f;
  
  if( g_bUseDiff )
  {
    float3 N = tex2Dlod( s1, float4( vVtUV, 0, 0 ) );
  
    float3 vLDir = float3( -0.577, -0.577, 0.577 );
    vLDir = mul( float4( vLDir, 0.0f ), g_mIW ).xyz;
    vDiff = max( 0, dot( N, -vLDir ) );
  }
  
  // Principal curvatures as color gradients
  if( iVisMode == 0 )
  {
    float fK1 = tex2Dlod( s2, float4( vVtUV, 0, 0 ) ).w;
    float fK2 = tex2Dlod( s3, float4( vVtUV, 0, 0 ) ).w;
	float H = 0.5f * (fK1 + fK2);
	float K = fK1 * fK2;
	float h = 4.0f / 3.0f * abs(atan2(H*H-K,H*H*sign(H)));
	float sc = M_2_PI * atan((2.0f*H*H-K) * SCALE * 2);
	
	vGD = float4( hsv2rgb( float3( h, sc, 1.0 ) ), 1.0f );
  }
  else 
  // Magnitude of principal curvatures in grayscale
  if( iVisMode == 1 )
  {
    float fK1 = tex2Dlod( s2, float4( vVtUV, 0, 0 ) ).w;
    float fK2 = tex2Dlod( s3, float4( vVtUV, 0, 0 ) ).w;
	float H = 0.5f * (fK1 + fK2);
	float c = (256.0f / M_PI * (atan(H*SCALE / 3) + M_PI_2)) / 255.0f;
	
	vGD = float4( c, c, c, 1.0f );
  }
  else
  // Gaussian curvature as colors
  if( iVisMode == 2 )
  {
    float fK1 = tex2Dlod( s2, float4( vVtUV, 0, 0 ) ).w;
    float fK2 = tex2Dlod( s3, float4( vVtUV, 0, 0 ) ).w;
	float K = ( fK1 * fK2 ) * ( SCALE / 2 );
	
	float s = 0.5f + K;
	
	vGD = float4( s, s, s, 1.0f );	
  }
  else
  // Median curvature as colors
  if( iVisMode == 3 )
  {
    float fK1 = tex2Dlod( s2, float4( vVtUV, 0, 0 ) ).w;
    float fK2 = tex2Dlod( s3, float4( vVtUV, 0, 0 ) ).w;
	float H = 0.5f * (fK1 + fK2) * ( SCALE / 15 );

	float s = 0.5f + H;
	
	vGD = float4( s, s, s, 1.0f );	
  }
  else
  // Magnitude of derivative of curvature
  if( iVisMode == 4 )
  {
	float4 vT3 = tex2Dlod( s4, float4( vVtUV, 0, 0 ) );
	float s = 1.0f;
	float c = vT3.x*vT3.x + vT3.y*vT3.y + vT3.z*vT3.z + vT3.w*vT3.w;
	c *= SCALE;
	
	if( c < 1.0f ) s -= 1.0f - c;
	if( c > M_PI ) c = M_PI;
	
	vGD = float4( hsv2rgb( float3( M_PI - c, s, 1.0 ) ), 1.0f );
  }
  
  Out.vColor = float4( vDiff, 1.0f ) * vGD;
 
  return Out;
}

////////////////////////////////////////////////////////////////////////////
// Pixel shader.
////////////////////////////////////////////////////////////////////////////
float4 RenderPS( VS_OUT In ) : COLOR
{
  return In.vColor;
}

////////////////////////////////////////////////////////////////////////////
// Effect techniques for rendering the actual geometry.
////////////////////////////////////////////////////////////////////////////

// Visualize principal curvatures as color gradients
technique Render0
{
  pass P0
  {
    VertexShader = compile vs_3_0 RenderVS( 0 );
    PixelShader  = compile ps_2_0 RenderPS( );
  }
}

// Visualize magnitude of principal curvatures in grayscale
technique Render1
{
  pass P0
  {
    VertexShader = compile vs_3_0 RenderVS( 1 );
    PixelShader  = compile ps_2_0 RenderPS( );
  }
}

// Visualize Gaussian curvature as colors
technique Render2
{
  pass P0
  {
    VertexShader = compile vs_3_0 RenderVS( 2 );
    PixelShader  = compile ps_2_0 RenderPS( );
  }
}

// Visualize mean curvature as colors
technique Render3
{
  pass P0
  {
    VertexShader = compile vs_3_0 RenderVS( 3 );
    PixelShader  = compile ps_2_0 RenderPS( );
  }
}

// Visualize magnitude of derivative of curvature
technique Render4
{
  pass P0
  {
    VertexShader = compile vs_3_0 RenderVS( 4 );
    PixelShader  = compile ps_2_0 RenderPS( );
  }
}

// Default visualization
technique Render5
{
  pass P0
  {
    VertexShader = compile vs_3_0 RenderVS( 5 );
    PixelShader  = compile ps_2_0 RenderPS( );
  }
}